#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np

from ciderpress import lib
from ciderpress.dft.lcao_convolutions import (
    ATCBasis,
    atom_loc_from_bas,
    get_gamma_lists_from_bas_and_env,
    libcider,
)
from ciderpress.dft.sph_harm_coeff import get_deriv_ylm_coeff

ATOM_OF = 0
ANG_OF = 1


def get_deriv_mol(old_bas, old_env):
    assert old_bas.ndim == 2 and old_bas.shape[1] == 8
    bas = []
    for b in old_bas:
        if b[ANG_OF] == 0:
            continue
        else:
            bas.append(b)
    bas = np.asarray(bas, order="C", dtype=np.int32)
    bas[:, ANG_OF] -= 1
    bas[:, 2:5] = 0
    loc = [0]
    for b in bas:
        loc.append(loc[-1] + 2 * b[ANG_OF] + 1)
    assert loc[-1] == np.sum(2 * bas[:, ANG_OF] + 1)
    loc = np.asarray(loc, order="C", dtype=np.int32)
    return bas, old_env.copy(), loc, atom_loc_from_bas(bas)


class LCAOInterpolator:
    """
    rads = aparam * (exp(dparam * arange(nrad)) - 1)
    """

    onsite_direct = False
    _ga_loc = None

    def __init__(
        self,
        atom_coords,
        atco,
        n0,
        n1,
        aparam=0.03,
        dparam=0.04,
        nrad=200,
    ):
        """
        Args:
            atom_coords (np.ndarray) : atomic coordinates (natm, 3)
            atco (ATCBasis) : atomic basis for interpolator. This is typically
                the atco_out object of a ConvolutionCollection
            n0 (int) : Number of l=0 features
            n1 (int) : Number of l=1 features
            aparam (float) : A parameter for radial grid.
            dparam (float) : D parameter for radial grid.
            nrad (int) : size of radial grid
        """
        assert atom_coords.shape == (atco.natm, 3)
        self.atom_coords = np.ascontiguousarray(atom_coords)
        self._aparam = aparam
        self._dparam = dparam
        self._nrad = nrad
        self.atco = atco
        lmax = np.max(atco.bas[:, ANG_OF])
        self.lmax = lmax
        self.nlm = (lmax + 1) * (lmax + 1)
        self.rads = np.asarray(
            aparam * (np.exp(dparam * np.arange(nrad)) - 1),
            dtype=np.float64,
            order="C",
        )
        self.w0_rsp = None
        self.wm_rsp = None
        self._n0 = n0
        self._n1 = n1
        self._gaunt_coeff = None
        self._l1bas = None
        self._l1loc = None
        self.l1atco = None

        self._l0bas = self.atco.bas
        self._l0env = self.atco.env
        self._l0loc = self.atco.ao_loc
        self._l0_atom_loc = self.atco.atom_loc
        self._ga_loc_ptr = lib.c_null_ptr()
        self._iatom_loc_ptr = lib.c_null_ptr()

        def _call_spline_(w_rsp, bas, env):
            nbas = len(bas)
            shls_slice = (0, nbas)
            libcider.compute_spline_maps(
                w_rsp.ctypes.data_as(ctypes.c_void_p),
                self.rads.ctypes.data_as(ctypes.c_void_p),
                bas.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(nbas),
                env.ctypes.data_as(ctypes.c_void_p),
                (ctypes.c_int * 2)(*shls_slice),
                ctypes.c_int(self.nrad),
            )

        shape = (nrad, self.atco.nbas, 4)
        if self._n0 > 0:
            self.w0_rsp = np.empty(shape, dtype=np.float64, order="C")
            _call_spline_(self.w0_rsp, self._l0bas, self._l0env)
        if self._n1 > 0:
            dbas, denv, dloc, aloc = get_deriv_mol(self._l0bas, self._l0env)
            dat = get_gamma_lists_from_bas_and_env(dbas, denv)
            self.l1atco = ATCBasis(*dat)
            self.wm_rsp = np.empty((nrad, len(dbas), 4), dtype=np.float64, order="C")
            self._gaunt_coeff = get_deriv_ylm_coeff(self.lmax)
            self._l1bas = dbas
            self._l1loc = dloc
            self._l1env = denv
            self._l1_atom_loc = aloc
            _call_spline_(self.wm_rsp, self._l1bas, self._l1env)

        self.is_num_ai_setup = False

    @property
    def num_in(self):
        """Get the number of (orbital-space) inputs (i.e. nbeta for ccl)"""
        return self._n0 + 2 * self._n1

    @property
    def num_out(self):
        """Get the number of (spline-space) outputs"""
        return self._n0 + 4 * self._n1

    @property
    def aparam(self):
        """Parameter a for radial spline grid (see __init__ docstring)."""
        return self._aparam

    @property
    def dparam(self):
        """Parameter d for radial spline grid (see __init__ docstring)."""
        return self._dparam

    @property
    def nrad(self):
        """Size of radial grid fpr spline"""
        return self._nrad

    @property
    def ngrids(self):
        if self.all_coords is None:
            raise RuntimeError("Uninitialized")
        else:
            return self.all_coords.shape[0]

    @classmethod
    def from_ccl(cls, atom_coords, ccl, aparam=0.03, dparam=0.04, nrad=200):
        """
        Initialize an instance of LCAOInterpolator from a
        ConvolutionCollection instance.

        Args:
            atom_coords (np.ndarray): (natm, 3) atomic coordinates
            ccl (ConvolutionCollection): ConvolutionCollection object
                from which to build the interpolator.
            aparam (float) : A parameter for radial grid.
            dparam (float) : D parameter for radial grid.
            nrad (int) : size of radial grid

        Returns:
            New LCAOInterpolator
        """
        return cls(
            atom_coords,
            ccl.atco_out,
            ccl.n0,
            ccl.n1,
            aparam=aparam,
            dparam=dparam,
            nrad=nrad,
        )

    def _orb2spline_(
        self, atco, f_arlpq, f_uq, w_rsp, nalpha, offset_spline, offset_orb, fwd
    ):
        """
        Helper function for converting between the orbital and spline bases.

        Args:
            f_arlpq (np.ndarray): Spline basis tensor
            f_uq (np.ndarray): Orbital basis tensor (in atco basis)
            w_rsp (np.ndarray): Splines for the orbitals
            offset_spline (int): q offset for the spline basis
            offset_orb (int): q offset for the orbital basis
            fwd (bool): fwd=True takes f_uq and writes f_arlpq.
                        fwd=False takes f_arlpq and writes f_uq.
        """
        assert f_arlpq.shape == (atco.natm, self.nrad, self.nlm, 4, self.num_out)
        assert f_uq.shape[0] == atco.nao
        assert w_rsp is not None
        assert f_arlpq.flags.c_contiguous
        assert f_uq.flags.c_contiguous
        assert w_rsp.flags.c_contiguous
        orb_stride = f_uq.shape[-1]
        spline_stride = f_arlpq.shape[-1]
        assert nalpha + offset_orb <= orb_stride
        assert nalpha + offset_spline <= spline_stride
        if fwd:
            fn = libcider.project_conv_to_spline
        else:
            fn = libcider.project_spline_to_conv
        fn(
            f_arlpq.ctypes.data_as(ctypes.c_void_p),
            f_uq.ctypes.data_as(ctypes.c_void_p),
            w_rsp.ctypes.data_as(ctypes.c_void_p),
            atco.atco_c_ptr,
            ctypes.c_int(nalpha),
            ctypes.c_int(self.nrad),
            ctypes.c_int(self.nlm),
            ctypes.c_int(orb_stride),
            ctypes.c_int(spline_stride),
            ctypes.c_int(offset_spline),
            ctypes.c_int(offset_orb),
        )

    def _fill_l1_coeff_(self, f_uq, f1_uq, offset, offset1, fwd):
        stride = f_uq.shape[1]
        stride1 = f1_uq.shape[1]
        if fwd:
            fn = libcider.fill_l1_coeff_fwd
        else:
            fn = libcider.fill_l1_coeff_bwd
        fn(
            f_uq.ctypes.data_as(ctypes.c_void_p),
            f1_uq.ctypes.data_as(ctypes.c_void_p),
            self._gaunt_coeff.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self.nlm),
            self.atco.atco_c_ptr,
            self.l1atco.atco_c_ptr,
            ctypes.c_int(stride),
            ctypes.c_int(offset),
            ctypes.c_int(stride1),
            ctypes.c_int(offset1),
        )

    def conv2spline(self, f_uq, f_arlpq=None, return_f1_uq=False):
        """
        Takes a set of convolutions stored in orbital space and transforms
        it into the spline basis

        Args:
            f_uq (np.ndarray): Orbital basis convolutions
            f_arlpq (np.ndarray): Spline basis convolutions

        Returns:
            f_arlpq (np.ndarray)
        """
        assert f_uq.flags.c_contiguous
        assert f_uq.shape == (self.atco.nao, self.num_in), (
            f_uq.shape,
            (self.atco.nao, self.num_in),
        )
        if f_arlpq is None:
            f_arlpq = np.zeros((self.atco.natm, self.nrad, self.nlm, 4, self.num_out))
        if self._n0 > 0:
            self._orb2spline_(
                self.atco, f_arlpq, f_uq, self.w0_rsp, self._n0, 0, 0, True
            )
        f1_uq = None
        if self._n1 > 0:
            f1_uq = np.zeros((self.l1atco.nao, 3 * self._n1))
            for i1 in range(self._n1):
                self._fill_l1_coeff_(f_uq, f1_uq, self._n0 + i1, 3 * i1, True)
            self._orb2spline_(
                self.l1atco,
                f_arlpq,
                f1_uq,
                self.wm_rsp,
                self._n1 * 3,
                self._n0,
                0,
                True,
            )
            offset_spline = self._n0 + 3 * self._n1
            offset_orb = self._n0 + self._n1
            self._orb2spline_(
                self.atco,
                f_arlpq,
                f_uq,
                self.w0_rsp,
                self._n1,
                offset_spline,
                offset_orb,
                True,
            )
        if return_f1_uq:
            return f_arlpq, f1_uq
        else:
            return f_arlpq

    def spline2conv(self, f_arlpq, f_uq=None, f1_uq=None):
        """
        Takes a set of convolutions stored in spline basis and transforms
        it into the orbital basis. Backwards version of spline2conv.

        Args:
            f_arlpq (np.ndarray): Spline basis convolutions
            f_uq (np.ndarray): Orbital basis convolutions

        Returns:
            f_uq (np.ndarray)
        """
        assert f_arlpq.flags.c_contiguous
        if f_uq is None:
            f_uq = np.zeros((self.atco.nao, self.num_in))
        assert f_uq.shape == (self.atco.nao, self.num_in)
        if self._n0 > 0:
            self._orb2spline_(
                self.atco, f_arlpq, f_uq, self.w0_rsp, self._n0, 0, 0, False
            )
        if self._n1 > 0:
            if f1_uq is None:
                f1_uq = np.zeros((self.l1atco.nao, 3 * self._n1))
            self._orb2spline_(
                self.l1atco,
                f_arlpq,
                f1_uq,
                self.wm_rsp,
                self._n1 * 3,
                self._n0,
                0,
                False,
            )
            for i1 in range(self._n1):
                self._fill_l1_coeff_(f_uq, f1_uq, self._n0 + i1, 3 * i1, False)
            offset_spline = self._n0 + 3 * self._n1
            offset_orb = self._n0 + self._n1
            self._orb2spline_(
                self.atco,
                f_arlpq,
                f_uq,
                self.w0_rsp,
                self._n1,
                offset_spline,
                offset_orb,
                False,
            )
        return f_uq

    @property
    def natm(self):
        return self.atco.natm

    def _set_num_ai(self, all_coords):
        if all_coords is None:
            assert self.is_num_ai_setup
            return
        all_coords = np.ascontiguousarray(all_coords)
        assert all_coords.shape[1] == 3
        ngrids_tot = all_coords.shape[0]
        self.num_ai = np.empty((self.natm, self.nrad), dtype=np.int32, order="C")
        """libcider.compute_num_spline_contribs(
            self.num_ai.ctypes.data_as(ctypes.c_void_p),
            all_coords.ctypes.data_as(ctypes.c_void_p),
            self.atom_coords.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_double(self.aparam),
            ctypes.c_double(self.dparam),
            ctypes.c_int(self.natm),
            ctypes.c_int(ngrids_tot),
            ctypes.c_int(self.nrad),
            self._ga_loc_ptr,
        )"""
        libcider.compute_num_spline_contribs_new(
            self.num_ai.ctypes.data_as(ctypes.c_void_p),
            all_coords.ctypes.data_as(ctypes.c_void_p),
            self.atom_coords.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_double(self.aparam),
            ctypes.c_double(self.dparam),
            ctypes.c_int(self.natm),
            ctypes.c_int(ngrids_tot),
            ctypes.c_int(self.nrad),
            self._iatom_loc_ptr,
        )
        self.all_coords = all_coords
        self._loc_ai = np.ascontiguousarray(
            (
                np.append(
                    np.zeros((self.natm, 1), dtype=np.int32),
                    np.cumsum(self.num_ai, axis=1),
                    axis=1,
                )
            ).astype(np.int32)
        )
        # buffers
        self._ind_ord_fwd = np.empty(ngrids_tot, dtype=np.int32, order="C")
        self._ind_ord_bwd = np.empty(ngrids_tot, dtype=np.int32, order="C")
        self._coords_ord = np.empty((ngrids_tot, 3), dtype=np.float64, order="C")
        # end buffers
        self._maxg = np.max(self.num_ai)

        self.is_num_ai_setup = True

    def _compute_spline_ind_order(self, a):
        if not self.is_num_ai_setup:
            raise RuntimeError
        ngrids_tot = self.all_coords.shape[0]
        nrad = self.num_ai.shape[1]
        assert self.all_coords.shape[1] == 3
        assert self.all_coords.flags.c_contiguous
        libcider.compute_spline_ind_order_new(
            self._loc_ai[a].ctypes.data_as(ctypes.c_void_p),
            self.all_coords.ctypes.data_as(ctypes.c_void_p),
            self.atom_coords[a].ctypes.data_as(ctypes.c_void_p),
            self._coords_ord.ctypes.data_as(ctypes.c_void_p),
            self._ind_ord_fwd.ctypes.data_as(ctypes.c_void_p),
            self._ind_ord_bwd.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_double(self.aparam),
            ctypes.c_double(self.dparam),
            ctypes.c_int(ngrids_tot),
            ctypes.c_int(nrad),
            self._iatom_loc_ptr,
            ctypes.c_int(a),
        )
        ngrids = np.max(self._loc_ai)
        nlm = self.nlm
        self._auxo_gl_buf = np.ndarray((ngrids, nlm), order="C")
        self._auxo_gp_buf = np.ndarray((ngrids, 4), order="C")
        return self._coords_ord, self._ind_ord_fwd

    def _eval_spline_bas_single(self, a):
        self._compute_spline_ind_order(a)
        ngrids = self._coords_ord.shape[0]
        if self.onsite_direct:
            # TODO maybe a bit inefficient, should clean this up
            # when I get around to parallelizing _compute_spline_ind_order
            ngrids -= np.sum(self.grids_indexer.iatom_list == a)
        nlm = self.nlm
        nrad = self.nrad
        atm_coord = self.atom_coords[a]
        auxo_gl = np.ndarray((ngrids, nlm), order="C", buffer=self._auxo_gl_buf)
        auxo_gp = np.ndarray((ngrids, 4), order="C", buffer=self._auxo_gp_buf)
        libcider.compute_spline_bas_separate(
            auxo_gl.ctypes.data_as(ctypes.c_void_p),
            auxo_gp.ctypes.data_as(ctypes.c_void_p),
            self._coords_ord.ctypes.data_as(ctypes.c_void_p),
            atm_coord.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(1),
            ctypes.c_int(ngrids),
            ctypes.c_int(nrad),
            ctypes.c_int(nlm),
            ctypes.c_double(self.aparam),
            ctypes.c_double(self.dparam),
        )
        return auxo_gl, auxo_gp

    def _eval_spline_bas_single_deriv(self, a):
        self._compute_spline_ind_order(a)
        ngrids = self._coords_ord.shape[0]
        if self.onsite_direct:
            # TODO maybe a bit inefficient, should clean this up
            # when I get around to parallelizing _compute_spline_ind_order
            ngrids -= np.sum(self.grids_indexer.iatom_list == a)
        nlm = self.nlm
        nrad = self.nrad
        atm_coord = self.atom_coords[a]
        auxo_vgl = np.ndarray((4, ngrids, nlm), order="C")
        auxo_vgp = np.ndarray((4, ngrids, 4), order="C")
        libcider.compute_spline_bas_separate_deriv(
            auxo_vgl.ctypes.data_as(ctypes.c_void_p),
            auxo_vgp.ctypes.data_as(ctypes.c_void_p),
            self._coords_ord.ctypes.data_as(ctypes.c_void_p),
            atm_coord.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(1),
            ctypes.c_int(ngrids),
            ctypes.c_int(nrad),
            ctypes.c_int(nlm),
            ctypes.c_double(self.aparam),
            ctypes.c_double(self.dparam),
        )
        return auxo_vgl, auxo_vgp

    def _call_l1_fill(self, f_gq, atom_coord, fwd):
        if fwd:
            l1_fn = libcider.add_lp1_term_fwd
        else:
            l1_fn = libcider.add_lp1_term_bwd
        for i1 in range(self._n1):
            ig = self.num_out + i1 - self._n1
            ix = self._n0 + 3 * i1
            l1_fn(
                f_gq.ctypes.data_as(ctypes.c_void_p),
                self.all_coords.ctypes.data_as(ctypes.c_void_p),
                atom_coord.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(f_gq.shape[0]),
                ctypes.c_int(ig),
                ctypes.c_int(ix + 0),
                ctypes.c_int(ix + 1),
                ctypes.c_int(ix + 2),
                ctypes.c_int(f_gq.shape[1]),
            )

    def _call_l1_fill_grad(self, excsum, f_gq, vf_gq, ia):
        natm = self.atco.natm
        iatom_list = self.grids_indexer.iatom_list
        n = iatom_list.size
        for i1 in range(self._n1):
            ig = self.num_out + i1 - self._n1
            ix = self._n0 + 3 * i1
            libcider.add_lp1_term_grad(
                excsum.ctypes.data_as(ctypes.c_void_p),
                f_gq.ctypes.data_as(ctypes.c_void_p),
                vf_gq.ctypes.data_as(ctypes.c_void_p),
                iatom_list.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(ia),
                ctypes.c_int(natm),
                ctypes.c_int(n),
                ctypes.c_int(ig),
                ctypes.c_int(ix + 0),
                ctypes.c_int(ix + 1),
                ctypes.c_int(ix + 2),
                ctypes.c_int(f_gq.shape[1]),
            )

    def _interpolate_nopar_atom(self, f_arlpq, f_gq, fwd):
        assert self.is_num_ai_setup
        ngrids_tot = self.all_coords.shape[0]
        nalpha = f_arlpq.shape[-1]
        assert f_arlpq.shape[-1] == f_gq.shape[-1]
        if fwd:
            fn = libcider.compute_mol_convs_single_new
        else:
            fn = libcider.compute_pot_convs_single_new
        for a in range(self.atco.natm):
            auxo_gl, auxo_gp = self._eval_spline_bas_single(a)
            if self.onsite_direct and len(self._loc_ai) == 1:
                if not fwd:
                    f_arlpq[:] = 0
            else:
                ngrids = ngrids_tot
                if self.onsite_direct:
                    ngrids -= self._ga_loc[a + 1] - self._ga_loc[a]
                if not fwd:
                    self._call_l1_fill(f_gq, self.atom_coords[a], fwd)
                fn(
                    f_gq.ctypes.data_as(ctypes.c_void_p),
                    f_arlpq[a].ctypes.data_as(ctypes.c_void_p),
                    auxo_gl.ctypes.data_as(ctypes.c_void_p),
                    auxo_gp.ctypes.data_as(ctypes.c_void_p),
                    self._loc_ai[a].ctypes.data_as(ctypes.c_void_p),
                    self._ind_ord_fwd.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(nalpha),
                    ctypes.c_int(self.nrad),
                    ctypes.c_int(ngrids),
                    ctypes.c_int(self.nlm),
                    ctypes.c_int(self._maxg),
                )
                if fwd:
                    self._call_l1_fill(f_gq, self.atom_coords[a], fwd)

    def _contract_grad_terms(self, excsum, f_g, a, v):
        iatom_list = self.grids_indexer.iatom_list
        assert iatom_list is not None
        assert iatom_list.flags.c_contiguous
        ngrids = iatom_list.size
        libcider.contract_grad_terms2(
            excsum.ctypes.data_as(ctypes.c_void_p),
            f_g.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(self.atco.natm),
            ctypes.c_int(a),
            ctypes.c_int(v),
            ctypes.c_int(ngrids),
            iatom_list.ctypes.data_as(ctypes.c_void_p),
        )

    def _interpolate_nopar_atom_deriv(self, f_arlpq, f_gq):
        assert self.is_num_ai_setup
        ngrids_tot = self.all_coords.shape[0]
        nalpha = f_arlpq.shape[-1]
        assert f_arlpq.shape[-1] == f_gq.shape[-1]
        fn = libcider.compute_mol_convs_single_new
        excsum = np.zeros((self.atco.natm, 3))
        ftmp_gq = np.empty_like(f_gq)
        for a in range(self.atco.natm):
            auxo_vgl, auxo_vgp = self._eval_spline_bas_single_deriv(a)
            if self.onsite_direct and len(self._loc_ai) == 1:
                pass
            else:
                ngrids = ngrids_tot
                if self.onsite_direct:
                    ngrids -= self._ga_loc[a + 1] - self._ga_loc[a]
                args = [
                    ftmp_gq.ctypes.data_as(ctypes.c_void_p),
                    f_arlpq[a].ctypes.data_as(ctypes.c_void_p),
                    auxo_vgl[0].ctypes.data_as(ctypes.c_void_p),
                    auxo_vgp[0].ctypes.data_as(ctypes.c_void_p),
                    self._loc_ai[a].ctypes.data_as(ctypes.c_void_p),
                    self._ind_ord_fwd.ctypes.data_as(ctypes.c_void_p),
                    ctypes.c_int(nalpha),
                    ctypes.c_int(self.nrad),
                    ctypes.c_int(ngrids),
                    ctypes.c_int(self.nlm),
                    ctypes.c_int(self._maxg),
                ]
                if self._n1 > 0:
                    ftmp_gq[:] = 0
                    fn(*args)
                    self._call_l1_fill_grad(excsum, ftmp_gq, f_gq, a)
                for v in range(3):
                    ftmp_gq[:] = 0
                    args[2] = auxo_vgl[0].ctypes.data_as(ctypes.c_void_p)
                    args[3] = auxo_vgp[v + 1].ctypes.data_as(ctypes.c_void_p)
                    fn(*args)
                    args[2] = auxo_vgl[v + 1].ctypes.data_as(ctypes.c_void_p)
                    args[3] = auxo_vgp[0].ctypes.data_as(ctypes.c_void_p)
                    fn(*args)
                    self._call_l1_fill(ftmp_gq, self.atom_coords[a], True)
                    # TODO accelerate since this step will be done many times
                    ftmp = np.einsum("gq,gq->g", ftmp_gq, f_gq)
                    self._contract_grad_terms(excsum, ftmp, a, v)
        return excsum

    def set_coords(self, coords):
        self._set_num_ai(coords)

    def interpolate_fwd(self, f_arlpq, f_gq=None):
        """

        Args:
            f_arlpq:
            f_gq: output
            atom_par (bool): If True, parallelize evaluation over atoms. Note,
                this can get quite memory intensive, so if atom_par=True,
                the user should only pass reasonably small sets of coords
                to this function.

        Returns:
            f_gq (np.ndarray): shape (coords.shape[0], f_arlpq.shape[-1])
        """
        if not self.is_num_ai_setup:
            raise ValueError("Need to set up indexes and coordinates first")
        assert self.all_coords is not None
        assert self.all_coords.ndim == 2
        assert self.all_coords.shape[1] == 3
        if f_gq is None:
            f_gq = np.zeros((self.all_coords.shape[0], self.num_out))
        self._interpolate_nopar_atom(f_arlpq, f_gq, True)
        return f_gq

    def interpolate_grad(self, f_arlpq, f_gq):
        if not self.is_num_ai_setup:
            raise ValueError("Need to set up indexes and coordinates first")
        assert self.all_coords is not None
        assert self.all_coords.ndim == 2
        assert self.all_coords.shape[1] == 3
        excsum = self._interpolate_nopar_atom_deriv(f_arlpq, f_gq)
        return excsum

    def interpolate_bwd(self, f_gq, f_arlpq=None):
        if not self.is_num_ai_setup:
            raise ValueError("Need to set up indexes and coordinates first")
        assert self.all_coords is not None
        assert self.all_coords.ndim == 2
        assert self.all_coords.shape[1] == 3
        if f_arlpq is None:
            f_arlpq = np.zeros((self.atco.natm, self.nrad, self.nlm, 4, self.num_out))
        self._interpolate_nopar_atom(f_arlpq, f_gq, False)
        return f_arlpq

    def project_orb2grid(self, f_uq, f_gq=None):
        """

        Args:
            f_uq:
            f_gq:
            spline_buf: Must be all zeros (TODO this is unsafe)

        Returns:

        """
        f_arlpq = self.conv2spline(f_uq)
        f_gq = self.interpolate_fwd(f_arlpq=f_arlpq, f_gq=f_gq)
        return f_gq

    def project_orb2grid_grad(self, f_uq, f_gq):
        f_arlpq = self.conv2spline(f_uq)
        excsum = self.interpolate_grad(f_arlpq=f_arlpq, f_gq=f_gq)
        return excsum

    def project_grid2orb(self, f_gq, f_uq=None):
        f_arlpq = self.interpolate_bwd(f_gq=f_gq)
        f_uq = self.spline2conv(f_arlpq, f_uq=f_uq)
        return f_uq


class LCAOInterpolatorDirect(LCAOInterpolator):
    def __init__(
        self,
        grids_indexer,
        atom_coords,
        atco,
        n0,
        n1,
        aparam=0.03,
        dparam=0.04,
        nrad=200,
        onsite_direct=True,
    ):
        """
        This modified version of LCAOInterpolator takes an AtomicGridsIndexer
        as an argument, which allows it to support computing onsite
        convolution terms directly rather than via spline.
        This can be done by calling interpolate_direct_fwd or
        interpolate_direct_bwd.

        Args:
            grids_indexer (AtomicGridsIndexer):
            atom_coords:
            atco:
            n0:
            n1:
            aparam:
            dparam:
            nrad:
            onsite_direct
        """
        super(LCAOInterpolatorDirect, self).__init__(
            atom_coords, atco, n0, n1, aparam=aparam, dparam=dparam, nrad=nrad
        )
        self.grids_indexer = grids_indexer
        self._ga_loc = np.ascontiguousarray(self.grids_indexer.ga_loc)
        self.onsite_direct = onsite_direct
        if self.onsite_direct:
            self._ga_loc_ptr = self._ga_loc.ctypes.data_as(ctypes.c_void_p)
            self._iatom_loc_ptr = self.grids_indexer.iatom_list.ctypes.data_as(
                ctypes.c_void_p
            )
        else:
            self._ga_loc_ptr = lib.c_null_ptr()
            self._iatom_loc_ptr = lib.c_null_ptr()

    @property
    def grid_loc_atom(self):
        return self._ga_loc.copy()

    def set_coords(self, coords):
        assert coords.shape[0] == (
            self.grids_indexer.idx_map.size + self.grids_indexer.padding
        )
        self._set_num_ai(coords[: self.grids_indexer.idx_map.size])

    def _run_onsite_orb2grid(self, atco, f_uq, f_gq, nalpha, offset1, offset2, fwd):
        f_rlmq = self.grids_indexer.empty_rlmq(nalpha)
        f_rlmq[:] = 0.0
        if fwd:
            atco.convert_rad2orb_(
                f_rlmq,
                f_uq,
                self.grids_indexer,
                self.grids_indexer.rad_arr,
                rad2orb=not fwd,
                offset=offset1,
            )
            self.grids_indexer.reduce_angc_ylm_(
                f_rlmq,
                f_gq,
                a2y=not fwd,
                offset=offset2,
            )
        else:
            self.grids_indexer.reduce_angc_ylm_(
                f_rlmq,
                f_gq,
                a2y=not fwd,
                offset=offset2,
            )
            atco.convert_rad2orb_(
                f_rlmq,
                f_uq,
                self.grids_indexer,
                self.grids_indexer.rad_arr,
                rad2orb=not fwd,
                offset=offset1,
            )

    def _run_onsite_lp1(self, f_gq, fwd):
        if fwd:
            fn = libcider.add_lp1_onsite_new_fwd
        else:
            fn = libcider.add_lp1_onsite_new_bwd
        for i1 in range(self._n1):
            ig = self.num_out + i1 - self._n1
            ix = self._n0 + 3 * i1
            nf = f_gq.shape[1]
            fn(
                f_gq.ctypes.data_as(ctypes.c_void_p),
                self.grids_indexer.rad_arr.ctypes.data_as(ctypes.c_void_p),
                self.grids_indexer.rad_loc.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(self.grids_indexer.nrad),
                self.grids_indexer.dirs.ctypes.data_as(ctypes.c_void_p),
                self.grids_indexer.ylm_loc.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(nf),
                ctypes.c_int(ig),
                ctypes.c_int(ix + 0),
                ctypes.c_int(ix + 1),
                ctypes.c_int(ix + 2),
            )

    def project_orb2grid(self, f_uq, f_gq=None):
        """

        Args:
            f_uq:
            f_gq:
            spline_buf: Must be all zeros (TODO this is unsafe)

        Returns:

        """
        ngpp = self.all_coords.shape[0] + self.grids_indexer.padding
        if f_gq is None:
            f_gq = np.zeros((ngpp, self.num_out))
        assert f_gq.shape == (ngpp, self.num_out)
        out_gq = f_gq[: self.all_coords.shape[0]]
        in_uq = f_uq
        assert out_gq.flags.c_contiguous
        if self.onsite_direct:
            f_arlpq, f1_uq = self.conv2spline(in_uq, return_f1_uq=True)
            # f_arlpq = self.conv2spline(in_uq, return_f1_uq=False)
            tmp_gq = np.zeros((self.grids_indexer.all_weights.size, self.num_out))
            # tmp_gq = np.zeros(
            #    (self.grids_indexer.all_weights.size, self._n0)
            # )
            self._run_onsite_orb2grid(self.atco, in_uq, tmp_gq, self._n0, 0, 0, True)
            if self._n1 > 0:
                self._run_onsite_orb2grid(
                    self.l1atco, f1_uq, tmp_gq, 3 * self._n1, 0, self._n0, True
                )
                o1 = self._n0 + self._n1
                o2 = self._n0 + 3 * self._n1
                self._run_onsite_orb2grid(
                    self.atco, in_uq, tmp_gq, self._n1, o1, o2, True
                )
                self._run_onsite_lp1(tmp_gq, True)
            out_gq[:] += tmp_gq[self.grids_indexer.idx_map]
        else:
            f_arlpq = self.conv2spline(in_uq, return_f1_uq=False)
        self.interpolate_fwd(f_arlpq=f_arlpq, f_gq=out_gq)
        return f_gq

    def project_grid2orb(self, f_gq, f_uq=None):
        ngpp = self.all_coords.shape[0] + self.grids_indexer.padding
        assert f_gq.shape == (ngpp, self.num_out)
        in_gq = f_gq[: self.all_coords.shape[0]]
        if f_uq is None:
            f_uq = np.zeros((self.atco.nao, self.num_in))
        out_uq = f_uq
        assert in_gq.flags.c_contiguous
        f_arlpq = self.interpolate_bwd(in_gq)
        if self.onsite_direct:
            tmp_gq = np.zeros((self.grids_indexer.all_weights.size, self.num_out))
            tmp_gq[self.grids_indexer.idx_map] = in_gq
            if self._n1 > 0:
                self._run_onsite_lp1(tmp_gq, False)
            self._run_onsite_orb2grid(self.atco, out_uq, tmp_gq, self._n0, 0, 0, False)
            if self._n1 > 0:
                f1_uq = np.zeros((self.l1atco.nao, 3 * self._n1))
                self._run_onsite_orb2grid(
                    self.l1atco, f1_uq, tmp_gq, 3 * self._n1, 0, self._n0, False
                )
                o1 = self._n0 + self._n1
                o2 = self._n0 + 3 * self._n1
                self._run_onsite_orb2grid(
                    self.atco, out_uq, tmp_gq, self._n1, o1, o2, False
                )
            else:
                f1_uq = None
            # f_arlpq = self.interpolate_bwd(in_gq)
            self.spline2conv(f_arlpq, f_uq=out_uq, f1_uq=f1_uq)
        else:
            # f_arlpq = self.interpolate_bwd(in_gq)
            self.spline2conv(f_arlpq, f_uq=out_uq)
        return out_uq

    @classmethod
    def from_ccl(
        cls,
        grids_indexer,
        atom_coords,
        ccl,
        aparam=0.03,
        dparam=0.04,
        nrad=200,
        onsite_direct=True,
    ):
        """
        Initialize an instance of LCAOInterpolator from a
        ConvolutionCollection instance.

        Args:
            atom_coords (np.ndarray): (natm, 3) atomic coordinates
            ccl (ConvolutionCollection): ConvolutionCollection object
                from which to build the interpolator.
            aparam (float) : A parameter for radial grid.
            dparam (float) : D parameter for radial grid.
            nrad (int) : size of radial grid
            onsite_direct (bool) :

        Returns:
            New LCAOInterpolator
        """
        return cls(
            grids_indexer,
            atom_coords,
            ccl.atco_out,
            ccl.n0,
            ccl.n1,
            aparam=aparam,
            dparam=dparam,
            nrad=nrad,
            onsite_direct=onsite_direct,
        )



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
from scipy.linalg import cholesky


def gauss_fft(n, a, k):
    # \int_0^\infty dr r^(2+n) exp(-ar^2) j_n(k*r)
    return (
        2 ** (-2 - n)
        * a ** (-1.5 - n)
        * np.exp(-(k**2) / (4 * a))
        * k**n
        * np.sqrt(np.pi)
    )


def gauss_and_derivs(a, r):
    r = np.sqrt(a) * r
    """
    E**(-r**2)
    (-2*r)/E**r**2
    -2/E**r**2 + (4*r**2)/E**r**2
    (12*r)/E**r**2 - (8*r**3)/E**r**2
    12/E**r**2 - (48*r**2)/E**r**2 + (16*r**4)/E**r**2
    """
    sqe = np.exp(-(r**2))
    vals = np.stack(
        [
            sqe,
            (-2 * r) * sqe,
            -2 * sqe + (4 * r**2) * sqe,
            (12 * r) * sqe - (8 * r**3) * sqe,
            12 * sqe - (48 * r**2) * sqe + (16 * r**4) * sqe,
        ]
    )
    for n in range(vals.shape[0]):
        vals[n] *= np.sqrt(a) ** n
    return vals


def get_ffunc(n, r, R):
    x = r / R
    xn = x**n
    y = xn * (x * x - 1) * (x * x - 1)
    y[x > 1] = 0.0
    return y


def get_poly(n, r, R):
    x = r / R
    xn = x**n
    xn[x > 1] = 0.0
    return xn


PFUNC2_NORMS = np.array(
    [
        5.77240697e00,
        3.07271389e00,
        3.14265261e00,
        2.13897490e01,
        1.94289398e01,
        1.33574495e02,
        1.14971278e02,
        7.98033330e02,
        6.65701240e02,
        4.65664626e03,
        3.80496685e03,
    ],
    dtype=np.float64,
)


def get_pfunc(n, r, R, enforce_cutoff=True):
    x = np.pi * r / R
    xp = r / R
    if n == 0:
        y = 0.5 + 0.5 * np.cos(x)
    elif n % 2 == 1:
        y1 = np.sin(0.5 * x) - 0.5 + 0.5 * np.cos(x)
        y1 *= 5
        y = xp ** (n - 1) * np.exp(-(n - 1) * xp**2) * y1
    else:
        y2 = 0.5 - 0.5 * np.cos(2 * x)
        y = xp ** (n - 2) * np.exp(-(n - 2) * xp**2) * y2
    if enforce_cutoff:
        y[r > R] = 0
    return y


def get_pfunc_norm(n, r, R, enforce_cutoff=True):
    if n >= len(PFUNC2_NORMS):
        raise ValueError
    res = get_pfunc(n, r, R, enforce_cutoff=True)
    res *= PFUNC2_NORMS[n] / R**1.5
    return res


def get_pfuncs_k(pfuncs_g, ls, rgd, ns=None):
    if ns is not None:
        l_add = ns - ls
    else:
        l_add = np.zeros_like(ls)
    n_pfuncs = pfuncs_g.shape[0]
    pfuncs_k = pfuncs_g.copy()
    for i in range(n_pfuncs):
        pfuncs_k[i] = rgd.transform_single_fwd(pfuncs_g[i], ls[i], l_add=l_add[i])
    return pfuncs_k


def get_phi_iabk(pfuncs_k, ks, alphas, betas=None):
    if betas is None:
        betas = alphas
    nalpha = alphas.shape[0]
    nbeta = betas.shape[0]
    n_pfuncs = pfuncs_k.shape[0]
    nk = ks.size
    phi_iabk = np.empty((n_pfuncs, nalpha, nbeta, nk))
    k2s = ks * ks
    for a in range(nalpha):
        for b in range(nbeta):
            aexp = alphas[a] + betas[b]
            expnt = 1 / (4 * aexp)
            prefac = (np.pi / aexp) ** 1.5
            kernel = prefac * np.exp(-expnt * k2s)
            for i in range(n_pfuncs):
                phi_iabk[i, a, b, :] = kernel * pfuncs_k[i]
    return phi_iabk


def get_phi_iabg(phi_iabk, ls, rgd):
    phi_iabg = phi_iabk.copy()
    n_pfuncs = phi_iabk.shape[0]
    nalpha = phi_iabk.shape[1]
    nbeta = phi_iabk.shape[2]
    for i in range(n_pfuncs):
        for a in range(nalpha):
            for b in range(nbeta):
                phi_iabg[i, a, b, :] = rgd.transform_single_bwd(
                    phi_iabk[i, a, b, :], ls[i]
                )
    return phi_iabg


def get_dv(rgd):
    return rgd.dr_g * rgd.r_g**2


def get_dvk(rgd):
    return 2 / np.pi * rgd.k_g**2 * rgd.dk_g


def get_p11_matrix(delta_l_pg, rgd, reg=0):
    dv_g = get_dv(rgd)
    p11_l_ii = []
    for l, delta_pg in enumerate(delta_l_pg):
        mat = np.einsum("pg,qg,g->pq", delta_pg, delta_pg, dv_g)
        mat[:] += reg * np.identity(mat.shape[-1])
        p11_l_ii.append(mat)
    return p11_l_ii


def get_p12_p21_matrix(delta_l_pg, phi_jabg, rgd, w_b, nbas_loc):
    dv_g = get_dv(rgd)
    p12_vbja = []
    p21_javb = []
    lp1 = len(delta_l_pg)
    for l in range(lp1):
        j0, j1 = nbas_loc[l], nbas_loc[l + 1]
        mat = np.einsum("vg,jabg,g->javb", delta_l_pg[l], phi_jabg[j0:j1], dv_g)
        mat[:] *= w_b
        p21_javb.append(mat)
        p12_vbja.append(mat.transpose(2, 3, 0, 1))
    return p12_vbja, p21_javb


def get_p22_matrix(phi_iabg, rgd, rcut, l, w_b, nbas_loc, reg=0, cut_func=False):
    dv_g = get_dv(rgd)
    phicut_iabg = phi_iabg[nbas_loc[l] : nbas_loc[l + 1]] * np.sqrt(dv_g)
    if cut_func:
        x = rgd.r_g / (0.75 * rcut)
        if cut_func == 2:
            fcut = np.ones_like(x)
        else:
            fcut = 0.5 + 0.5 * np.cos(np.pi * x)
        fcut[x > 1] = 0.0
        phicut_iabg *= np.sqrt(fcut)
    p22_jaja = np.einsum("iacg,jbcg,c->iajb", phicut_iabg, phicut_iabg, w_b)
    nj, na = p22_jaja.shape[:2]
    p22_ii = p22_jaja.view()
    p22_ii.shape = (nj * na, nj * na)
    # TODO this is closed to previous regularization, but not
    # necessarily the optimal regularization
    reg = reg / np.tile(w_b, nj)
    p22_ii[:] += reg * np.identity(nj * na)
    return p22_jaja


def construct_full_p_matrices(p11_l_vv, p12_l_vbja, p21_l_javb, p22_l_jaja, w_b):
    lp1 = len(p11_l_vv)
    p_l_ii = []
    for l in range(lp1):
        nv = p11_l_vv[l].shape[0]
        nb = w_b.size
        N1 = nv * nb
        p11_pbpb = np.zeros((nv, nb, nv, nb))
        for b in range(nb):
            p11_pbpb[:, b, :, b] = p11_l_vv[l] * w_b[b]
        p11_ii = p11_pbpb.reshape(N1, N1)
        p12_vbja = p12_l_vbja[l]
        p21_javb = p21_l_javb[l]
        p22_jaja = p22_l_jaja[l]
        N2 = p22_jaja.shape[0] * p22_jaja.shape[1]
        N = N1 + N2
        pl_ii = np.zeros((N, N))
        pl_ii[:N1, :N1] = p11_ii
        pl_ii[:N1, N1:] = p12_vbja.reshape(N1, N2)
        pl_ii[N1:, :N1] = p21_javb.reshape(N2, N1)
        pl_ii[N1:, N1:] = p22_jaja.reshape(N2, N2)
        p_l_ii.append(pl_ii)
    return p_l_ii


def get_delta_lpg(betas_lb, rcut, rgd, pmin, pmax):
    r_g = rgd.r_g
    dv_g = get_dv(rgd)
    lp1 = len(betas_lb)
    delta_lpg = []
    for l in range(lp1):
        cond = np.logical_and(
            betas_lb[l] > pmin,
            betas_lb[l] < pmax,
        )
        betas = betas_lb[l][cond]
        fcut = (0.5 * np.pi / betas) ** 0.75 * gauss_and_derivs(betas, rcut)
        funcs = (0.5 * np.pi / betas[:, None]) ** 0.75 * np.exp(
            -betas[:, None] * r_g * r_g
        )
        rd_g = rgd.r_g - rcut
        poly = (
            fcut[0, :, None]
            + rd_g * fcut[1, :, None]
            + rd_g**2 * fcut[2, :, None] / 2
            + rd_g**3 * fcut[3, :, None] / 6
        )
        dpoly = fcut[1] - rcut * fcut[2] + rcut**2 * fcut[3] / 2
        poly += dpoly[:, None] / (4 * rcut**3) * rd_g**4
        funcs = funcs - poly
        funcs[:, rgd.r_g > rcut] = 0
        funcs_tmp = funcs * r_g**l * np.sqrt(dv_g)
        ovlp = np.einsum("ig,jg->ij", funcs_tmp, funcs_tmp)
        L = cholesky(ovlp, lower=True)
        basis = np.linalg.solve(L, funcs)
        delta_lpg.append(basis * r_g**l)
    return delta_lpg



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from abc import ABC, abstractmethod

import numpy as np
from pyscf import dft, gto, lib, scf
from pyscf.dft.gen_grid import Grids
from pyscf.sgx.sgx import sgx_fit

from ciderpress.external.sgx_tools import get_jk_densities

CALC_TYPES = {
    "RHF": scf.hf.RHF,
    "UHF": scf.uhf.UHF,
}


def recursive_remove_none(obj):
    if isinstance(obj, dict):
        return {k: recursive_remove_none(v) for k, v in obj.items() if v is not None}
    else:
        return obj


def recursive_bytes_to_str(obj):
    if isinstance(obj, dict):
        return {k: recursive_bytes_to_str(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        lst = [recursive_bytes_to_str(item) for item in obj]
        if isinstance(obj, tuple):
            lst = tuple(lst)
        return lst
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        return obj


class ElectronAnalyzer(ABC):
    """
    A class for generating and storing data derived from a PySCF
    electronic structure calculation, in particular distributions
    on a real-space grid such as the density, exchange (correlation)
    energy density, Coulomb energy density, etc.

    Attributes:

        TODO

    """

    _atype = None

    def __init__(
        self, mol, dm, grids_level=3, mo_occ=None, mo_coeff=None, mo_energy=None
    ):
        if mo_occ is None and hasattr(dm, "mo_occ"):
            self.mo_occ = dm.mo_occ
        else:
            self.mo_occ = mo_occ
        if mo_coeff is None and hasattr(dm, "mo_coeff"):
            self.mo_coeff = dm.mo_coeff
        else:
            self.mo_coeff = mo_coeff
        if mo_energy is None and hasattr(dm, "mo_energy"):
            self.mo_energy = dm.mo_energy
        else:
            self.mo_energy = mo_energy
        self.dm = dm
        self.mol = mol
        self.mol.verbose = 3
        self.mol.build()
        self.grids_level = grids_level
        self.grids = Grids(self.mol)
        self.grids.level = self.grids_level
        self._data = {}
        self.grids.build(with_non0tab=False)

    def keys(self):
        return self._data.keys()

    def set(self, k, v):
        self._data[k] = v

    def as_dict(self):
        return {
            "atype": self._atype,
            "data": self._data,
            "dm": self.dm,
            "mo_occ": self.mo_occ,
            "mo_coeff": self.mo_coeff,
            "mo_energy": self.mo_energy,
            "grids_level": self.grids_level,
            "mol": gto.mole.pack(self.mol),
        }

    @staticmethod
    def from_dict(d):
        args = [
            gto.mole.unpack(d["mol"]),
            d["dm"],
        ]
        kwargs = {
            "grids_level": d["grids_level"],
            "mo_occ": d.get("mo_occ"),
            "mo_coeff": d.get("mo_coeff"),
            "mo_energy": d.get("mo_energy"),
        }
        atype = d.get("atype")
        if atype is None:
            cls = ElectronAnalyzer
        elif atype == "RHF" or atype == "RKS":
            cls = RHFAnalyzer
        elif atype == "UHF" or atype == "UKS":
            cls = UHFAnalyzer
        else:
            raise ValueError("Unrecognized analyzer type {}".format(atype))
        analyzer = cls(*args, **kwargs)
        analyzer._data.update(d["data"])
        return analyzer

    def dump(self, fname):
        """
        Dump self to an hdf5 file called fname.

        Args:
            fname (str): Name of file to dump to
        """
        h5dict = recursive_remove_none(self.as_dict())
        lib.chkfile.dump(fname, "analyzer", h5dict)

    @staticmethod
    def load(fname):
        """
        Load instance of cls from hdf5
        Args:
            fname (str): Name of file from which to load
            max_mem: See __init__
        """
        analyzer_dict = lib.chkfile.load(fname, "analyzer")
        analyzer_dict = recursive_bytes_to_str(analyzer_dict)
        return ElectronAnalyzer.from_dict(analyzer_dict)

    def get(self, name):
        return self._data[name]

    def calculate_vxc(self, xcname, xcfunc=None, grids=None, xctype="MGGA"):
        """
        Computes the XC potential matrix for a functional,
        store it, and return it.

        Args:
            xcname (str): XC functional
            xcfunc (callable): XC function in case one
                wants to use an XC functional not in libxc
            grids (optional): Grid to use to calculate XC.

        Returns:
            vxc (nspin x nao x nao): XC potential matrix
        """
        if not isinstance(xcname, str):
            raise ValueError("XC must be string")
        if self._atype == "RHF":
            calc = dft.RKS(self.mol)
        elif self._atype == "UHF":
            calc = dft.UKS(self.mol)
        else:
            raise NotImplementedError
        calc.xc = xcname
        if grids is not None:
            calc.grids = grids
        if xcfunc is not None:
            calc.define_xc_(xcfunc, xctype=xctype)
        vxc = calc.get_veff(dm=self.rdm1)
        self._data["EXC_{}".format(xcname)] = vxc.exc
        vxc = vxc - vxc.vj
        self._data["VXC_{}".format(xcname)] = vxc
        return vxc

    def calculate_vxc_on_mo(self, xcname, orbs=None, **kwargs):
        """
        Compute contributions of XC potential to the eigenvalues. If
        VXC_{xcname} is not in _data, calculate_vxc is called first.

        Args:
            xcname (str): Name of XC functional (for libxc)
            orbs: Orbital index dictionary for computing orbital derivatives,
                in the format use for descriptors.get_descriptors.
            **kwargs: extra arguments to pass to calculate_vxc, if it is called.

        Returns:
            eigenvalue contributions. If orbs is provided, it will be in
                the same format used by descritpors.get_descriptors.
                Otherwise, it will be a numpy array with all the eigenvalue
                contributions.
        """
        if "VXC_{}".format(xcname) not in self._data.keys():
            vxc = self.calculate_vxc(xcname, **kwargs)
        else:
            vxc = self._data["VXC_{}".format(xcname)]
        assert self.mo_coeff is not None
        if orbs is None:
            orbvals = vxc.dot(self.mo_coeff)
            eigvals = np.einsum("...ui,ui->...i", self.mo_coeff, orbvals)
            self._data["ORBXC_{}".format(xcname)] = eigvals
        else:
            from ciderpress.pyscf.descriptors import (
                get_labels_and_coeffs,
                unpack_eigvals,
            )

            labels, coeffs, en_list, sep_spins = get_labels_and_coeffs(
                orbs, self.mo_coeff, self.mo_occ, self.mo_energy
            )
            if self.mo_occ.ndim == 2:
                orbvals = {}
                for s in range(2):
                    c = np.array(coeffs[s])
                    if len(c) == 0:
                        orbvals[s] = None
                        continue
                    orbvals[s] = np.array(c).dot(vxc[s])
                    orbvals[s] = np.einsum("iu,iu->i", c, orbvals[s])
            else:
                c = np.array(coeffs)
                orbvals = c.dot(vxc)
                orbvals = np.einsum("iu,iu->i", c, orbvals)
            if sep_spins:
                eigvals = (
                    unpack_eigvals(orbs[0], labels[0], orbvals[0]),
                    unpack_eigvals(orbs[1], labels[1], orbvals[1]),
                )
            elif self.mo_occ.ndim == 2:
                eigvals = {}
                for k in list(orbs.keys()):
                    eigvals[k] = {}
                for s in range(2):
                    eigvals_tmp = unpack_eigvals(orbs, labels[s], orbvals[s])
                    for k in list(orbs.keys()):
                        eigvals[k].update(eigvals_tmp[k])
            else:
                eigvals = unpack_eigvals(orbs, labels, orbvals)
            self._data["ORBXC_{}".format(xcname)] = eigvals
        return self._data["ORBXC_{}".format(xcname)]

    @staticmethod
    def from_calc(calc, grids_level=None, store_energy_orig=True):
        """
        NOTE: This has side effects on calc, see notes below.

        Args:
            calc: SCF object from PySCF
            grids_level: Size of PySCF grids for XC integration
            store_energy_orig: Whether to store original xc energy.

        Returns:

        """
        if isinstance(calc, scf.uhf.UHF):
            cls = UHFAnalyzer
        elif isinstance(calc, scf.hf.RHF):
            cls = RHFAnalyzer
        else:
            raise ValueError("Must be HF calculation")

        if grids_level is None:
            if isinstance(calc, dft.rks.KohnShamDFT):
                grids_level = calc.grids.level
            else:
                grids_level = 3
        analyzer = cls(
            calc.mol,
            calc.make_rdm1(),
            grids_level=grids_level,
            mo_occ=calc.mo_occ,
            mo_coeff=calc.mo_coeff,
            mo_energy=calc.mo_energy,
        )
        if store_energy_orig and isinstance(calc, dft.rks.KohnShamDFT):
            # save initial exc for reference
            if (calc.scf_summary.get("exc") is None) or (
                grids_level != calc.grids.level
            ):
                # NOTE this should leave calc almost the same
                # as it started, but might make some
                # minor changes if there are non-default
                # settings/screenings on the grids. Also
                # scf_summary will be overwritten
                old_level = calc.grids.level
                calc.grids.level = grids_level
                calc.grids.build()
                if hasattr(calc, "with_df") and hasattr(calc.with_df, "grids"):
                    calc.with_df.build()
                e_tot = calc.energy_tot(analyzer.dm)
                calc.grids.level = old_level
                calc.grids.build()
            else:
                e_tot = calc.e_tot
            analyzer._data["xc_orig"] = calc.xc
            analyzer._data["exc_orig"] = calc.scf_summary["exc"]
            analyzer._data["e_tot_orig"] = e_tot
        return analyzer

    def get_rho_data(self, overwrite=False):
        if "rho_data" in self.keys() and not overwrite:
            return self.get("rho_data")
        ni = dft.numint.NumInt()
        spinpol = self.dm.ndim == 3
        if spinpol:
            rho_list_a = []
            rho_list_b = []
        else:
            rho_list = []
        for ao, non0, weight, coord in ni.block_loop(self.mol, self.grids, deriv=2):
            if spinpol:
                rho_list_a.append(ni.eval_rho(self.mol, ao, self.dm[0], xctype="MGGA"))
                rho_list_b.append(ni.eval_rho(self.mol, ao, self.dm[1], xctype="MGGA"))
            else:
                rho_list.append(ni.eval_rho(self.mol, ao, self.dm, xctype="MGGA"))
        if spinpol:
            rhoa = np.concatenate(rho_list_a, axis=-1)
            rhob = np.concatenate(rho_list_b, axis=-1)
            self._data["rho_data"] = np.stack([rhoa, rhob], axis=0)
        else:
            self._data["rho_data"] = np.concatenate(rho_list, axis=-1)
        return self._data["rho_data"]

    @abstractmethod
    def perform_full_analysis(self):
        pass

    @property
    def rdm1(self):
        return self.dm

    @property
    def rho_data(self):
        return self.get("rho_data")


class RHFAnalyzer(ElectronAnalyzer):

    _atype = "RHF"

    def get_xc(self, xcname):
        """
        Function for accessing XC functional data from _data
        """
        if xcname in self._data["xc_list"]:
            return self._data["xc"][xcname]
        else:
            raise ValueError("Data not computed for xc functional")

    def get_rs(self, term, omega, tol=1e-9):
        """
        Function for accessing range-separated data from _data
        Finds match based on a tolerance, in case rounding
        issues arise with floats.
        """
        if not (term in ["ee", "ha", "ex"]):
            raise ValueError("Term must be ee, ha, or ex")
        name = "{}_energy_density_rs".format(term)
        for ind, omega_tst in enumerate(self._data["omega_list"]):
            if abs(omega_tst - omega) < tol:
                return self._data[name][ind]
        else:
            raise RuntimeError("RS data not found for omega={}".format(omega))

    def get_xc_energy(self, xcname):
        """
        Store and return the XC energy for the functional given by xcname
        """
        if self._data.get("xc") is None:
            self._data["xc_list"] = []
            self._data["xc"] = {}
        ni = dft.numint.NumInt()
        if isinstance(self, UHFAnalyzer):
            _, exc, _ = ni.nr_uks(
                self.mol, self.grids, xcname, self.dm, relativity=0, hermi=1
            )
        else:
            _, exc, _ = ni.nr_rks(
                self.mol, self.grids, xcname, self.dm, relativity=0, hermi=1
            )
        self._data["xc_list"].append(xcname)
        self._data["xc"][xcname] = exc
        return exc

    def get_ha_energy_density(self):
        self.get_ee_energy_density()
        return self._data["ha_energy_density"]

    def get_ex_energy_density(self):
        self.get_ee_energy_density()
        return self._data["ex_energy_density"]

    get_fx_energy_density = get_ex_energy_density

    def _get_ej_ek(self, omega=None):
        calc_tmp = sgx_fit(scf.RHF(self.mol))
        calc_tmp.build()
        calc_tmp.with_df.grids = self.grids
        if omega is None:
            ej, ek = get_jk_densities(calc_tmp.with_df, self.dm)
        else:
            with calc_tmp.mol.with_range_coulomb(omega):
                ej, ek = get_jk_densities(calc_tmp.with_df, self.dm)
        return ej, ek

    def _get_ee(self, omega=None):
        ej, ek = self._get_ej_ek(omega)
        ej = ej[0]
        ek = 0.5 * ek[0]
        return ej, ek, ej + ek

    def get_ee_energy_density(self):
        """
        Returns the sum of E_{Ha} and E_{X}, i.e. the Coulomb repulsion
        energy of the HF or KS Slater determinant.
        """
        ej, ek, ee = self._get_ee()
        self._data["ha_energy_density"] = ej
        self._data["ex_energy_density"] = ek
        self._data["ee_energy_density"] = ee
        return self._data["ee_energy_density"]

    def get_ee_energy_density_rs(self, omega):
        if self._data.get("omega_list") is None:
            self._data["omega_list"] = []
            self._data["ee_energy_density_rs"] = []
            self._data["ha_energy_density_rs"] = []
            self._data["ex_energy_density_rs"] = []
        ej, ek, ee = self._get_ee(omega)
        self._data["omega_list"].append(omega)
        self._data["ha_energy_density_rs"].append(ej)
        self._data["ex_energy_density_rs"].append(ek)
        self._data["ee_energy_density_rs"].append(ee)
        return self._data["ee_energy_density_rs"][-1]

    def perform_full_analysis(self):
        self.get_rho_data()
        self.get_ee_energy_density()

    @property
    def atype(self):
        return self._atype


RKSAnalyzer = RHFAnalyzer


class UHFAnalyzer(RHFAnalyzer):

    _atype = "UHF"

    def _get_ej_ek(self, omega=None):
        calc_tmp = sgx_fit(scf.UHF(self.mol))
        calc_tmp.build()
        calc_tmp.with_df.grids = self.grids
        if omega is None:
            ej, ek = get_jk_densities(calc_tmp.with_df, self.dm)
        else:
            with calc_tmp.mol.with_range_coulomb(omega):
                ej, ek = get_jk_densities(calc_tmp.with_df, self.dm)
        return ej, ek

    def _get_ee(self, omega=None):
        ej, ek = self._get_ej_ek(omega)
        return ej, ek, ej + ek


UKSAnalyzer = UHFAnalyzer



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import numpy as np
import torch
import torch.nn as nn


class XNet(nn.Module):
    def __init__(self, ninp, nlayer=5, nhidden=30):
        super(XNet, self).__init__()
        n = nhidden
        self.l0 = nn.Linear(ninp, n)
        self.lls = nn.ModuleList([nn.Linear(n, n) for i in range(nlayer)])
        self.l2s = nn.ModuleList([nn.Linear(ninp, n) for i in range(nlayer)])
        self.last = nn.Linear(n, 1)
        self.af = nn.Tanh()

    def forward(self, x):
        x0 = x
        x = self.l0(x)
        for ll, l2 in zip(self.lls, self.l2s):
            x = l2(x0) * x + ll(x)
            x = self.af(x)
        return torch.squeeze(self.last(x))


class XNet2(nn.Module):
    def __init__(self, ninp, nlayer=5, nhidden=30):
        super(XNet2, self).__init__()
        n = nhidden
        self.l0 = nn.Linear(ninp, n)
        self.lls = nn.ModuleList([nn.Linear(n, n) for i in range(nlayer)])
        self.l2s = nn.ModuleList([nn.Linear(ninp, n) for i in range(nlayer)])
        self.last = nn.Linear(n, 1)
        self.af = nn.Tanh()

    def forward(self, x):
        x0 = x
        x = self.l0(x)
        for ll, l2 in zip(self.lls, self.l2s):
            x = l2(x0) + ll(x) + x
            x = self.af(x)
        return torch.squeeze(self.last(x))


class XNet3(nn.Module):
    def __init__(self, ninp, nlayer=5, nhidden=30):
        super(XNet3, self).__init__()
        n = nhidden
        self.l0 = nn.Linear(ninp, n)
        self.lls = nn.ModuleList([nn.Linear(n, n) for i in range(nlayer)])
        self.l2s = nn.ModuleList([nn.Linear(ninp, n) for i in range(nlayer)])
        self.last = nn.Linear(n, 1)
        self.af = nn.Tanh()

    def forward(self, x):
        x0 = x
        x = self.l0(x)
        for ll, l2 in zip(self.lls, self.l2s):
            x = l2(x0) + ll(x)
            x = self.af(x)
        return torch.squeeze(self.last(x))


def simple_map(
    kernel, use_grad=True, nlayer=5, nhidden=30, nstep=3000, nncls=XNet, lr=1.0
):
    if torch.cuda.is_available():
        dev = "cuda:0"
    else:
        dev = "cpu"
    device = torch.device(dev)

    # gp = joblib.load(fname)
    # X = gp.kernels[0].X1ctrl
    X = kernel.X1ctrl
    alpha = kernel.alpha
    # alpha = gp.kernels[0].alpha
    # k = gp.kernels[0].kernel(X, X)
    k = kernel.kernel(X, X)
    y = alpha.dot(k)
    Xtr = X.copy()

    def eval_gp(Xtst):
        k = kernel.kernel(Xtr, Xtst)
        return alpha.dot(k)

    my_nn = nncls(X.shape[1], nlayer=nlayer, nhidden=nhidden)
    my_nn = my_nn.double().to(device)
    print(my_nn)

    loss = nn.MSELoss()
    # optimizer = torch.optim.SGD(my_nn.parameters(), lr=1e-2)
    optimizer = torch.optim.LBFGS(my_nn.parameters(), lr=lr)

    def train(X, y, model, loss_fn, optimizer, t, need_closure=False):
        size = len(X)
        model.train()
        pred = model(X)
        loss = loss_fn(pred, y)
        loss.backward()

        def closure_in():
            optimizer.zero_grad()
            pred = model(X)
            loss = loss_fn(pred, y)
            loss.backward()
            return loss

        if need_closure:
            optimizer.step(closure_in)
        else:
            optimizer.step()
        optimizer.zero_grad()
        loss = loss.item()
        if t % 10 == 0:
            print(f"loss: {loss:>7f}  [{t:>5d}/{size:>5d}]")

    def train_grad(
        X, y, model, loss_fn, optimizer, t, delta, nfeat, nsamp, need_closure=False
    ):
        size = len(X)
        model.train()
        pred = model(X)
        fac = 1.0 / nfeat
        loss = loss_fn(pred[:nsamp], y[:nsamp])
        for i in range(nfeat):
            deriv = (
                pred[(2 * i + 1) * nsamp : (2 * i + 2) * nsamp]
                - pred[(2 * i + 2) * nsamp : (2 * i + 3) * nsamp]
            ) / delta
            loss = loss + fac * loss_fn(deriv, y[(i + 1) * nsamp : (i + 2) * nsamp])
        loss.backward()

        def closure_in():
            optimizer.zero_grad()
            pred = model(X)
            loss = loss_fn(pred[:nsamp], y[:nsamp])
            for i in range(nfeat):
                deriv = (
                    pred[(2 * i + 1) * nsamp : (2 * i + 2) * nsamp]
                    - pred[(2 * i + 2) * nsamp : (2 * i + 3) * nsamp]
                ) / delta
                loss = loss + fac * loss_fn(deriv, y[(i + 1) * nsamp : (i + 2) * nsamp])
            loss.backward()
            return loss

        if need_closure:
            optimizer.step(closure_in)
        else:
            optimizer.step()
        optimizer.zero_grad()
        loss = loss.item(), 0
        if t % 10 == 0:
            print(f"loss: {loss:>7f}  [{t:>5d}/{size:>5d}]")

    delta = 1e-4
    nsamp = X.shape[0]
    Xlst = [X]
    ylst = [y]
    for i in range(X.shape[-1]):
        Xp = X.copy()
        Xp[:, i] += 0.5 * delta
        Xm = X.copy()
        Xm[:, i] -= 0.5 * delta
        Xlst.append(Xp)
        Xlst.append(Xm)
        dy = (eval_gp(Xp) - eval_gp(Xm)) / delta
        print(dy.shape)
        ylst.append(dy)
    X = np.concatenate(Xlst, axis=0)
    y = np.concatenate(ylst)
    X = torch.tensor(X).to(device)
    y = torch.tensor(y).to(device)
    for t in range(nstep):
        if use_grad:
            train_grad(
                X,
                y,
                my_nn,
                loss,
                optimizer,
                t,
                delta,
                X.shape[1],
                nsamp,
                need_closure=True,
            )
        else:
            train(X, y, my_nn, loss, optimizer, t, need_closure=True)
    my_nn.eval()
    import time

    t0 = time.monotonic()
    for i in range(100):
        pred = my_nn(X[:nsamp])
    t1 = time.monotonic()
    print(t1 - t0)
    diff = (y[:nsamp] - pred).cpu().detach().numpy()
    print(np.mean(np.abs(diff)))
    print(np.sqrt(np.mean(diff**2)))
    print(np.max(np.abs(diff)))
    print("Done")
    return my_nn.to(torch.device("cpu"))



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes

import numpy as np
from gpaw.xc.gga import GGA
from gpaw.xc.mgga import MGGA

from ciderpress.gpaw.atom_utils import FastAtomPASDWSlice, FastPASDWCiderKernel
from ciderpress.gpaw.cider_fft import (
    CIDERPW_GRAD_MODE_FORCE,
    CIDERPW_GRAD_MODE_NONE,
    CIDERPW_GRAD_MODE_STRESS,
    CiderGGA,
    CiderMGGA,
    add_gradient_correction,
)
from ciderpress.gpaw.nldf_interface import pwutil as libpwutil


class CiderPASDW_MPRoutines:
    has_paw = True

    def get_D_asp(self):
        return self.atomdist.to_work(self.dens.D_asp)

    def add_forces(self, F_av):
        nt_sg = self.dens.nt_xg
        redist = self._hamiltonian.xc_redistributor
        if redist is not None:
            nt_sg = redist.distribute(nt_sg)
        self.calculate_force_contribs(F_av, nt_sg)

    def _collect_paw_corrections(self):
        """
        For use in xc_tools, only after calculate_paw_correction
        is called. Collects potential using from_work
        """
        dH_asp_tmp = self.dH_asp_tmp
        E_a_tmp = self.E_a_tmp
        dH_asp_new = self.dens.setups.empty_atomic_matrix(
            self.nspin, self.atom_partition
        )
        E_a_new = self.atom_partition.arraydict([(1,)] * len(self.dens.setups), float)
        for a, E in E_a_tmp.items():
            E_a_new[a][:] = E
            dH_asp_new[a][:] = dH_asp_tmp[a]
        dist = self.atomdist
        self.E_a_tmp = dist.from_work(E_a_new)
        self.dH_asp_tmp = dist.from_work(dH_asp_new)

    def initialize_paw_kernel(self, cider_kernel_inp, Nalpha_atom, encut_atom):
        self.paw_kernel = FastPASDWCiderKernel(
            cider_kernel_inp,
            self._plan,
            self.aux_gd,
            self.cut_xcgrid,
        )
        self.paw_kernel.initialize(
            self.dens, self.atomdist, self.atom_partition, self.setups
        )
        self.paw_kernel.initialize_more_things(self.setups)
        self.paw_kernel.interpolate_dn1 = self.interpolate_dn1
        self.atom_slices_s = None

    def _setup_atom_slices(self):
        self.timer.start("ATOM SLICE SETUP")
        self.atom_slices_s = {}
        self.atom_slices_a = {}
        self.global_slices_s = {}
        self.global_slices_a = {}
        rank_a = self.atom_partition.rank_a
        comm = self.atom_partition.comm
        for a in range(len(self.setups)):
            args = [
                self.aux_gd,
                self.spos_ac[a],
                self.setups[a].ps_setup,
                self.fft_obj,
            ]
            kwargs = dict(
                rmax=self.setups[a].ps_setup.rcut_func,
                sphere=True,
                store_funcs=self.pasdw_store_funcs,
            )
            self.atom_slices_s[a] = FastAtomPASDWSlice.from_gd_and_setup(
                *args, **kwargs
            )
            if rank_a[a] == comm.rank and self.pasdw_ovlp_fit:
                self.global_slices_s[a] = FastAtomPASDWSlice.from_gd_and_setup(
                    *args, is_global=True, ovlp_fit=True, **kwargs
                )
            args[2] = self.setups[a].pa_setup
            kwargs["rmax"] = self.setups[a].pa_setup.rcut_func
            self.atom_slices_a[a] = FastAtomPASDWSlice.from_gd_and_setup(
                *args, **kwargs
            )
            if rank_a[a] == comm.rank and self.pasdw_ovlp_fit:
                self.global_slices_a[a] = FastAtomPASDWSlice.from_gd_and_setup(
                    *args, is_global=True, ovlp_fit=True, **kwargs
                )
        self.timer.stop()

    def write_augfeat_to_rbuf(self, s, pot=False):
        self.timer.start("PAW TO GRID")
        if self.atom_slices_s is None:
            self._setup_atom_slices()

        self.timer.start("setup")
        if pot:
            atom_slices = self.atom_slices_a
            global_slices = self.global_slices_a
        else:
            atom_slices = self.atom_slices_s
            global_slices = self.global_slices_s

        shapes = [
            (atom_slices[a].num_funcs, self._plan.nalpha)
            for a in range(len(self.setups))
        ]
        sizes = np.array([s[0] * s[1] for s in shapes])
        rank_a = self.atom_partition.rank_a
        comm = self.atom_partition.comm
        if self.fft_obj.has_mpi:
            atoms = [[] for r in range(comm.size)]
            for a in range(len(self.setups)):
                atoms[rank_a[a]].append(a)
            alocs = {}
            rlocs = []
            rsizes = []
            tot = 0
            for r in range(comm.size):
                rlocs.append(tot)
                for a in atoms[r]:
                    alocs[a] = tot
                    tot += sizes[a]
                rsizes.append(tot - rlocs[-1])
            sendbuf = np.empty(rsizes[comm.rank], dtype=np.float64)
            rtot = 0
            for a in atoms[comm.rank]:
                # assert c_abi[a].flags.c_contiguous
                c_iq = self.c_asiq[a][s]
                if self.pasdw_ovlp_fit:
                    sendbuf[rtot : rtot + sizes[a]] = np.ascontiguousarray(
                        # c_abi[a].dot(global_slices[a].sinv_pf.T)
                        global_slices[a].sinv_pf.dot(c_iq)
                    ).ravel()
                else:
                    sendbuf[rtot : rtot + sizes[a]] = np.ascontiguousarray(c_iq).ravel()
                rtot += sizes[a]
            np.asarray([rtot] * comm.size)
            np.asarray([0] * comm.size)
            recvbuf = np.empty(tot, dtype=np.float64)
            self.timer.stop()
            self.timer.start("comm")
            rsizes = np.asarray(rsizes, order="C", dtype=np.int32)
            rlocs = np.asarray(rlocs, order="C", dtype=np.int32)
            libpwutil.ciderpw_all_gatherv_from_gpaw(
                ctypes.py_object(comm.get_c_object()),
                sendbuf.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(rtot),
                recvbuf.ctypes.data_as(ctypes.c_void_p),
                rsizes.ctypes.data_as(ctypes.c_void_p),
                rlocs.ctypes.data_as(ctypes.c_void_p),
            )
        else:
            max_size = np.max([s[0] * s[1] for s in shapes])
            buf = np.empty(max_size, dtype=np.float64)
        self.timer.stop()

        for a in range(len(self.setups)):
            atom_slice = atom_slices[a]
            if self.fft_obj.has_mpi:
                coefs_iq = np.ndarray(shapes[a], buffer=recvbuf[alocs[a] :])
            else:
                coefs_iq = np.ndarray(shapes[a], buffer=buf)
                coefs_iq = np.ndarray(shapes[a], buffer=buf)
                if comm.rank == rank_a[a]:
                    coefs_iq[:] = self.c_asiq[a][s]
                comm.broadcast(coefs_iq, rank_a[a])
            if atom_slice is not None and atom_slices[a].rad_g.size > 0:
                self.timer.start("funcs")
                funcs_ig = atom_slice.get_funcs()
                funcs_gq = np.dot(funcs_ig.T, coefs_iq)
                self.timer.stop()
                self.timer.start("paw2grid")
                self.fft_obj.add_paw2grid(funcs_gq, atom_slice.indset)
                self.timer.stop()
        self.timer.stop()

    def interpolate_rbuf_onto_atoms(self, s, pot=False):
        self.timer.start("GRID TO PAW")
        if self.atom_slices_s is None:
            self._setup_atom_slices()

        if pot:
            atom_slices = self.atom_slices_s
            global_slices = self.global_slices_s
        else:
            atom_slices = self.atom_slices_a
            global_slices = self.global_slices_a

        shapes = [
            (atom_slices[a].num_funcs, self._plan.nalpha)
            for a in range(len(self.setups))
        ]

        rank_a = self.atom_partition.rank_a
        comm = self.atom_partition.comm
        coefs_a_iq = {}
        for a in range(len(self.setups)):
            atom_slice: FastAtomPASDWSlice = atom_slices[a]
            if atom_slice.rad_g.size > 0:
                funcs_ig = atom_slice.get_funcs()
                funcs_gq = np.empty((funcs_ig.shape[1], self._plan.nalpha))
                self.fft_obj.set_grid2paw(
                    funcs_gq,
                    atom_slice.indset,
                )
                coefs_iq = np.dot(funcs_ig, funcs_gq)
                coefs_iq[:] *= atom_slice.dv
            else:
                coefs_iq = np.zeros(shapes[a])
            coefs_a_iq[a] = coefs_iq
        for a in range(len(self.setups)):
            comm.sum(coefs_a_iq[a], rank_a[a])
        for a in range(len(self.setups)):
            coefs_iq = coefs_a_iq[a]
            if rank_a[a] == comm.rank:
                assert coefs_iq is not None
                if self.pasdw_ovlp_fit:
                    self.c_asiq[a][s] = global_slices[a].sinv_pf.T.dot(coefs_iq)
                else:
                    self.c_asiq[a][s] = coefs_iq
        self.timer.stop()

    def interpolate_rbuf_onto_atoms_grad(self, s, pot=False):
        self.timer.start("GRID TO PAW GRAD")
        if self.atom_slices_s is None:
            self._setup_atom_slices()

        if pot:
            atom_slices = self.atom_slices_s
            global_slices = self.global_slices_s
        else:
            atom_slices = self.atom_slices_a
            global_slices = self.global_slices_a

        shapes = [
            (atom_slices[a].num_funcs, self._plan.nalpha)
            for a in range(len(self.setups))
        ]

        rank_a = self.atom_partition.rank_a
        comm = self.atom_partition.comm
        coefs_a_viq = {}
        for a in range(len(self.setups)):
            atom_slice: FastAtomPASDWSlice = atom_slices[a]
            if atom_slice.rad_g.size > 0:
                funcs_vig = atom_slice.get_grads()
                funcs_gq = np.empty((funcs_vig.shape[2], self._plan.nalpha))
                self.fft_obj.set_grid2paw(
                    funcs_gq,
                    atom_slice.indset,
                )
                coefs_viq = np.dot(funcs_vig, funcs_gq)
                coefs_viq[:] *= atom_slice.dv
            else:
                coefs_viq = np.zeros((3,) + shapes[a])
            coefs_a_viq[a] = coefs_viq
        for a in range(len(self.setups)):
            comm.sum(coefs_a_viq[a], rank_a[a])
        for a in range(len(self.setups)):
            coefs_viq = coefs_a_viq[a]
            if rank_a[a] == comm.rank:
                assert coefs_viq is not None
                for v in range(3):
                    if self.pasdw_ovlp_fit:
                        self._save_v_avsiq[a][v, s] = global_slices[a].sinv_pf.T.dot(
                            coefs_viq[v]
                        )
                    else:
                        self._save_v_avsiq[a][v, s] = coefs_viq[v]
                if self.pasdw_ovlp_fit:
                    sl = global_slices[a]
                    X_vii = sl.get_ovlp_deriv(
                        sl.get_funcs(), sl.get_grads(), stress=False
                    )
                    # TODO not sure if this line below is quite right but it seems to work
                    # could also just be self.c_asiq[a][s]
                    # but in principle we need to remove the sinv_pf from c_asiq before multiplying
                    # in the derivative matrix
                    coefs_iq = np.linalg.solve(sl.sinv_pf.T, self.c_asiq[a][s])
                    self._save_v_avsiq[a][:, s] += np.einsum(
                        "vij,iq->vjq", X_vii, coefs_iq
                    )
        self.timer.stop()

    def interpolate_rbuf_onto_atoms_stress(self, s, pot=False):
        self.timer.start("GRID TO PAW GRAD")
        if self.atom_slices_s is None:
            self._setup_atom_slices()

        if pot:
            atom_slices = self.atom_slices_s
            global_slices = self.global_slices_s
        else:
            atom_slices = self.atom_slices_a
            global_slices = self.global_slices_a

        shapes = [
            (atom_slices[a].num_funcs, self._plan.nalpha)
            for a in range(len(self.setups))
        ]

        rank_a = self.atom_partition.rank_a
        comm = self.atom_partition.comm
        coefs_a_vviq = {}
        for a in range(len(self.setups)):
            atom_slice: FastAtomPASDWSlice = atom_slices[a]
            if atom_slice.rad_g.size > 0:
                funcs_vvig = atom_slice.get_stress_funcs()
                funcs_gq = np.empty((funcs_vvig.shape[-1], self._plan.nalpha))
                self.fft_obj.set_grid2paw(
                    funcs_gq,
                    atom_slice.indset,
                )
                coefs_vviq = np.dot(funcs_vvig, funcs_gq)
                coefs_vviq[:] *= atom_slice.dv
            else:
                coefs_vviq = np.zeros((3, 3) + shapes[a])
            coefs_a_vviq[a] = coefs_vviq
        for a in range(len(self.setups)):
            comm.sum(coefs_a_vviq[a], rank_a[a])
        for a in range(len(self.setups)):
            coefs_vviq = coefs_a_vviq[a]
            if rank_a[a] == comm.rank:
                assert coefs_vviq is not None
                for u in range(3):
                    for v in range(3):
                        if self.pasdw_ovlp_fit:
                            self._save_v_avsiq[a][u, v, s] = global_slices[
                                a
                            ].sinv_pf.T.dot(coefs_vviq[u, v])
                        else:
                            self._save_v_avsiq[a][u, v, s] = coefs_vviq[u, v]
                if self.pasdw_ovlp_fit:
                    sl = global_slices[a]
                    X_vvii = sl.get_ovlp_deriv(
                        sl.get_funcs(), sl.get_grads(), stress=True
                    )
                    coefs_iq = np.linalg.solve(sl.sinv_pf.T, self.c_asiq[a][s])
                    self._save_v_avsiq[a][:, :, s] += np.einsum(
                        "uvij,iq->uvjq", X_vvii, coefs_iq
                    )
        self.timer.stop()

    def set_fft_work(self, s, pot=False):
        self.write_augfeat_to_rbuf(s, pot=pot)

    def get_fft_work(self, s, pot=False, grad_mode=CIDERPW_GRAD_MODE_NONE):
        self.interpolate_rbuf_onto_atoms(s, pot=pot)
        if grad_mode == CIDERPW_GRAD_MODE_FORCE:
            if s == 0:
                self._save_v_avsiq = {}
                for a, c_siq in self.c_asiq.items():
                    self._save_v_avsiq[a] = np.zeros((3,) + c_siq.shape)
            self.interpolate_rbuf_onto_atoms_grad(s, pot=pot)
        elif grad_mode == CIDERPW_GRAD_MODE_STRESS:
            if s == 0:
                self._save_v_avsiq = {}
                for a, c_siq in self.c_asiq.items():
                    self._save_v_avsiq[a] = np.zeros((3, 3) + c_siq.shape)
            self.interpolate_rbuf_onto_atoms_stress(s, pot=pot)

    def _set_paw_terms(self):
        self.timer.start("PAW TERMS")
        D_asp = self.get_D_asp()
        if not (D_asp.partition.rank_a == self.atom_partition.rank_a).all():
            raise ValueError("rank_a mismatch")
        self.c_asiq, self.df_asgLq = self.paw_kernel.calculate_paw_feat_corrections(
            self.setups, D_asp
        )
        self._save_c_asiq = {k: v.copy() for k, v in self.c_asiq.items()}
        self.D_asp = D_asp
        self.timer.stop()

    def _calculate_paw_energy_and_potential(self, grad_mode=CIDERPW_GRAD_MODE_NONE):
        self.timer.start("PAW ENERGY")
        if grad_mode == CIDERPW_GRAD_MODE_STRESS:
            ctmp_asiq = {a: c_siq.copy() for a, c_siq in self.c_asiq.items()}
        self.c_asiq, deltaE, deltaV = self.paw_kernel.calculate_paw_feat_corrections(
            self.setups, self.D_asp, D_asiq=self.c_asiq, df_asgLq=self.df_asgLq
        )
        self.E_a_tmp = deltaE
        self.deltaV = deltaV
        self.timer.stop()
        if grad_mode == CIDERPW_GRAD_MODE_STRESS:
            tot = 0
            for a, ctmp_siq in ctmp_asiq.items():
                tot += np.einsum("siq,siq->", self.c_asiq[a], ctmp_siq)
            return tot

    def _get_dH_asp(self):
        self.timer.start("PAW POTENTIAL")
        dH_asp = self.paw_kernel.calculate_paw_feat_corrections(
            self.setups, self.D_asp, vc_asiq=self.c_asiq
        )
        for a in self.D_asp.keys():
            dH_asp[a] += self.deltaV[a]
        self.dH_asp_tmp = dH_asp
        self.timer.stop()

    def calculate_paw_correction(
        self, setup, D_sp, dEdD_sp=None, addcoredensity=True, a=None
    ):
        if a is None:
            raise ValueError("Atom index a not provided.")
        if dEdD_sp is not None:
            dEdD_sp += self.dH_asp_tmp[a]
        return self.E_a_tmp[a]

    def initialize_more_things(self):
        self.setups = self._hamiltonian.setups
        self.atomdist = self._hamiltonian.atomdist
        self.atom_partition = self.get_D_asp().partition

        encut_atom = self.encut
        Nalpha_atom = self.Nalpha
        while encut_atom < self.encut_atom_min:
            encut_atom *= self.lambd
            Nalpha_atom += 1

        self.initialize_paw_kernel(self.cider_kernel, Nalpha_atom, encut_atom)

    @property
    def is_initialized(self):
        if (
            (self._plan is None)
            or (self._plan.nspin != self.nspin)
            or (self.setups is None)
            or (self.atom_slices_s is None)
        ):
            return False
        else:
            return True


class CiderGGAPASDW(CiderPASDW_MPRoutines, CiderGGA):
    def __init__(self, cider_kernel, **kwargs):
        CiderGGA.__init__(
            self,
            cider_kernel,
            **kwargs,
        )
        defaults_list = {
            "interpolate_dn1": False,  # bool
            "encut_atom_min": 8000.0,  # float
            "cut_xcgrid": False,  # bool
            "pasdw_ovlp_fit": True,
            "pasdw_store_funcs": False,
        }
        settings = {}
        for k, v in defaults_list.items():
            settings[k] = kwargs.get(k)
            if settings[k] is None:
                settings[k] = defaults_list[k]
        self.__dict__.update(settings)
        self.k2_k = None
        self.setups = None

    def todict(self):
        d = super(CiderGGAPASDW, self).todict()
        d["_cider_type"] = "CiderGGAPASDW"
        d["xc_params"]["pasdw_store_funcs"] = self.pasdw_store_funcs
        d["xc_params"]["pasdw_ovlp_fit"] = self.pasdw_ovlp_fit
        return d

    def add_forces(self, F_av):
        GGA.add_forces(self, F_av)
        CiderPASDW_MPRoutines.add_forces(self, F_av)

    def stress_tensor_contribution(self, n_sg):
        return CiderGGA.stress_tensor_contribution(self, n_sg)

    def set_positions(self, spos_ac):
        self.spos_ac = spos_ac
        self.atom_slices_s = None

    def initialize(self, density, hamiltonian, wfs):
        self.dens = density
        self._hamiltonian = hamiltonian
        self.timer = wfs.timer
        self.world = wfs.world
        self._plan = None
        self.setups = None
        self._check_parallelization(wfs)

    def calculate_force_contribs(self, F_av, n_sg):
        e_g = np.zeros(n_sg.shape[1:])
        Ftmp_av = np.zeros_like(F_av)
        taut_sg, dedtaut_sg, taut_sG = self._get_taut(n_sg)
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, taut_sg)
        e_g[:] = 0.0
        fs = self.calc_cider(_e, rho_sxg, dedrho_sxg, grad_mode=CIDERPW_GRAD_MODE_FORCE)
        for a, f in fs.items():
            Ftmp_av[a] = f
        self.world.sum(Ftmp_av, 0)
        if self.world.rank == 0:
            F_av[:] += Ftmp_av


class CiderMGGAPASDW(CiderPASDW_MPRoutines, CiderMGGA):
    def __init__(self, cider_kernel, **kwargs):
        CiderMGGA.__init__(
            self,
            cider_kernel,
            **kwargs,
        )
        defaults_list = {
            "interpolate_dn1": False,  # bool
            "encut_atom_min": 8000.0,  # float
            "cut_xcgrid": False,  # bool
            "pasdw_ovlp_fit": True,
            "pasdw_store_funcs": False,
        }
        settings = {}
        for k, v in defaults_list.items():
            settings[k] = kwargs.get(k)
            if settings[k] is None:
                settings[k] = defaults_list[k]
        self.__dict__.update(settings)
        self.k2_k = None
        self.setups = None

    def todict(self):
        d = super(CiderMGGAPASDW, self).todict()
        d["_cider_type"] = "CiderMGGAPASDW"
        d["xc_params"]["pasdw_store_funcs"] = self.pasdw_store_funcs
        d["xc_params"]["pasdw_ovlp_fit"] = self.pasdw_ovlp_fit
        return d

    def add_forces(self, F_av):
        MGGA.add_forces(self, F_av)
        CiderPASDW_MPRoutines.add_forces(self, F_av)

    def set_positions(self, spos_ac):
        MGGA.set_positions(self, spos_ac)
        self.spos_ac = spos_ac
        self.atom_slices_s = None

    def initialize(self, density, hamiltonian, wfs):
        MGGA.initialize(self, density, hamiltonian, wfs)
        self.dens = density
        self._hamiltonian = hamiltonian
        self.timer = wfs.timer
        self.world = wfs.world
        self.setups = None
        self.atom_slices_s = None
        self._plan = None
        self._check_parallelization(wfs)

    def _get_taut(self, n_sg):
        self.timer.start("KED SETUP")
        if self.type == "MGGA":
            if self.fixed_ke:
                taut_sG = self._fixed_taut_sG
                if not self._taut_gradv_init:
                    self._taut_gradv_init = True
                    # ensure initialization for calculation potential
                    self.wfs.calculate_kinetic_energy_density()
            else:
                taut_sG = self.wfs.calculate_kinetic_energy_density()

            if taut_sG is None:
                taut_sG = self.wfs.gd.zeros(len(n_sg))

            taut_sg = np.empty_like(n_sg)

            for taut_G, taut_g in zip(taut_sG, taut_sg):
                taut_G += 1.0 / self.wfs.nspins * self.tauct_G
                self.distribute_and_interpolate(taut_G, taut_g)

            dedtaut_sg = np.zeros_like(n_sg)
        else:
            taut_sg = None
            dedtaut_sg = None
            taut_sG = None
        self.timer.stop()
        return taut_sg, dedtaut_sg, taut_sG

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        taut_sg, dedtaut_sg, taut_sG = self._get_taut(n_sg)
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, taut_sg)
        tmp_dedrho_sxg = np.zeros((len(n_sg), rho_sxg.shape[1]) + e_g.shape)
        e_g[:] = 0.0
        self.calc_cider(_e, rho_sxg, dedrho_sxg)
        self._add_from_cider_grid(e_g[None, :], _e[None, :])
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_dedrho_sxg[s], dedrho_sxg[s])
        v_sg[:] = tmp_dedrho_sxg[:, 0]
        add_gradient_correction(self.grad_v, tmp_dedrho_sxg, v_sg)

        if dedtaut_sg is not None:
            dedtaut_sg[:] = tmp_dedrho_sxg[:, 4]
            self.dedtaut_sG = self.wfs.gd.empty(self.wfs.nspins)
            self.ekin = 0.0
            for s in range(self.wfs.nspins):
                self.restrict_and_collect(dedtaut_sg[s], self.dedtaut_sG[s])
                self.ekin -= self.wfs.gd.integrate(
                    self.dedtaut_sG[s] * (taut_sG[s] - self.tauct_G / self.wfs.nspins)
                )

    def calculate_force_contribs(self, F_av, n_sg):
        e_g = np.zeros(n_sg.shape[1:])
        Ftmp_av = np.zeros_like(F_av)
        taut_sg, dedtaut_sg, taut_sG = self._get_taut(n_sg)
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, taut_sg)
        e_g[:] = 0.0
        fs = self.calc_cider(_e, rho_sxg, dedrho_sxg, grad_mode=CIDERPW_GRAD_MODE_FORCE)
        for a, f in fs.items():
            Ftmp_av[a] = f
        self.world.sum(Ftmp_av, 0)
        if self.world.rank == 0:
            F_av[:] += Ftmp_av

    def stress_tensor_contribution(self, n_sg):
        e_g = np.zeros(n_sg.shape[1:])
        stress1_vv = np.zeros((3, 3))
        taut_sg, dedtaut_sg, taut_sG = self._get_taut(n_sg)
        _e, rho_sxg, dedrho_sxg = self._get_cider_inputs(n_sg, taut_sg)
        tmp_dedrho_sxg = np.zeros((len(n_sg), rho_sxg.shape[1]) + e_g.shape)
        e_g[:] = 0.0
        fs = self.calc_cider(
            _e, rho_sxg, dedrho_sxg, grad_mode=CIDERPW_GRAD_MODE_STRESS
        )
        stress1_vv[:] += 0.5 * (fs + fs.T)
        self.world.sum(stress1_vv, 0)

        self._add_from_cider_grid(e_g[None, :], _e[None, :])
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_dedrho_sxg[s], dedrho_sxg[s])
        dedrho_sxg = tmp_dedrho_sxg
        tmp_rho_sxg = np.zeros_like(tmp_dedrho_sxg)
        for s in range(len(n_sg)):
            self._add_from_cider_grid(tmp_rho_sxg[s], rho_sxg[s])
        rho_sxg = tmp_rho_sxg

        def integrate(a1_g, a2_g=None):
            return self.gd.integrate(a1_g, a2_g, global_integral=False)

        P = integrate(e_g)
        nspin = n_sg.shape[0]
        for s in range(nspin):
            for x in range(5):
                P -= integrate(dedrho_sxg[s, x], rho_sxg[s, x])

        dedtaut_sg[:] = tmp_dedrho_sxg[:, 4]
        tau_svvG = self.wfs.calculate_kinetic_energy_density_crossterms()

        stress_vv = P * np.eye(3)
        tau_cross_g = self.gd.empty()
        for s in range(nspin):
            for v1 in range(3):
                for v2 in range(3):
                    stress_vv[v1, v2] -= integrate(
                        rho_sxg[s, v1 + 1], dedrho_sxg[s, v2 + 1]
                    )
                    self.distribute_and_interpolate(tau_svvG[s, v1, v2], tau_cross_g)
                    stress_vv[v1, v2] -= integrate(tau_cross_g, dedtaut_sg[s])

        self.dedtaut_sG = self.wfs.gd.empty(self.wfs.nspins)
        for s in range(self.wfs.nspins):
            self.restrict_and_collect(dedtaut_sg[s], self.dedtaut_sG[s])

        stress_vv = 0.5 * (stress_vv + stress_vv.T)

        self.gd.comm.sum(stress_vv)

        stress_vv += stress1_vv

        self.gd.comm.broadcast(stress_vv, 0)
        return stress_vv



import joblib
import yaml

from ciderpress.dft.xc_evaluator import MappedXC
from ciderpress.dft.xc_evaluator2 import MappedXC2


def load_cider_model(mlfunc, mlfunc_format):
    if isinstance(mlfunc, str):
        if mlfunc_format is None:
            if mlfunc.endswith(".yaml"):
                mlfunc_format = "yaml"
            elif mlfunc.endswith(".joblib"):
                mlfunc_format = "joblib"
            else:
                raise ValueError("Unsupported file format")
        if mlfunc_format == "yaml":
            with open(mlfunc, "r") as f:
                mlfunc = yaml.load(f, Loader=yaml.CLoader)
        elif mlfunc_format == "joblib":
            mlfunc = joblib.load(mlfunc)
        else:
            raise ValueError("Unsupported file format")
    if not isinstance(mlfunc, (MappedXC, MappedXC2)):
        raise ValueError("mlfunc must be MappedXC")
    return mlfunc


def get_slxc_settings(xc, xkernel, ckernel, xmix, cider_x_only=True):
    if xc is None:
        # xc is another way to specify non-mixed part of kernel
        xc = ""
    if ckernel is not None:
        if cider_x_only:
            xc = ckernel + " + " + xc
        else:
            xc = "{} * {} + {}".format(1 - xmix, ckernel, xc)
    if xkernel is not None:
        xc = "{} * {} + {}".format(1 - xmix, xkernel, xc)
    if xc.endswith(" + "):
        xc = xc[:-3]
    return xc



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

from abc import ABC, abstractmethod

import numpy as np
import yaml
from interpolation.splines.eval_cubic_numba import (
    vec_eval_cubic_splines_1,
    vec_eval_cubic_splines_2,
    vec_eval_cubic_splines_3,
    vec_eval_cubic_splines_4,
    vec_eval_cubic_splines_G_1,
    vec_eval_cubic_splines_G_2,
    vec_eval_cubic_splines_G_3,
    vec_eval_cubic_splines_G_4,
)


def get_vec_eval(grid, coeffs, X, N):
    """
    Call the numba-accelerated spline evaluation routines from the
    interpolation package. Also returns derivatives
    Args:
        grid: start and end points + number of grids in each dimension
        coeffs: coefficients of the spline
        X: coordinates to interpolate
        N: dimension of the interpolation (between 1 and 4, inclusive)
    """
    coeffs = np.expand_dims(coeffs, coeffs.ndim)
    y = np.zeros((X.shape[0], 1))
    dy = np.zeros((X.shape[0], N, 1))
    a_, b_, orders = zip(*grid)
    if N == 1:
        vec_eval_cubic_splines_G_1(a_, b_, orders, coeffs, X, y, dy)
    elif N == 2:
        vec_eval_cubic_splines_G_2(a_, b_, orders, coeffs, X, y, dy)
    elif N == 3:
        vec_eval_cubic_splines_G_3(a_, b_, orders, coeffs, X, y, dy)
    elif N == 4:
        vec_eval_cubic_splines_G_4(a_, b_, orders, coeffs, X, y, dy)
    else:
        raise ValueError("invalid dimension N")
    return np.squeeze(y, -1), np.squeeze(dy, -1)


def get_cubic(grid, coeffs, X, N):
    """
    Call the numba-accelerated spline evaluation routines from the
    interpolation package.
    Args:
        grid: start and end points + number of grids in each dimension
        coeffs: coefficients of the spline
        X: coordinates to interpolate
        N: dimension of the interpolation (between 1 and 4, inclusive)
    """
    coeffs = np.expand_dims(coeffs, coeffs.ndim)
    y = np.zeros((X.shape[0], 1))
    a_, b_, orders = zip(*grid)
    if N == 1:
        vec_eval_cubic_splines_1(a_, b_, orders, coeffs, X, y)
    elif N == 2:
        vec_eval_cubic_splines_2(a_, b_, orders, coeffs, X, y)
    elif N == 3:
        vec_eval_cubic_splines_3(a_, b_, orders, coeffs, X, y)
    elif N == 4:
        vec_eval_cubic_splines_4(a_, b_, orders, coeffs, X, y)
    else:
        raise ValueError("invalid dimension N")
    return np.squeeze(y, -1)


class Evaluator:
    """
    The Evaluator module computes the exchange energy densities
    and exchange enhancement factors for spline-interpolated XC functionals.
    """

    def __init__(
        self,
        scale,
        ind_sets,
        spline_grids,
        coeff_sets,
        xed_y_converter,
        feature_list,
        desc_order,
        const=0,
        desc_version="c",
        a0=8.0,
        fac_mul=0.25,
        amin=0.05,
        args=None,
    ):
        """
        Args:
            scale: lists of weights for the spline functions
            ind_sets: List of index sets. For each set of indices, passes
                the features at those indices to the spline interpolation
                functions.
            spline_grids: grids on which the splines are evaluated
            coeff_sets: Coefficients for the spline interpolation
            xed_y_converter (tuple): set of functions for convert between the
                exchange energy density (xed) and exchange enhancement
                factor y
                    (xed_to_y, y_to_xed, eval_baseline_y,
                    1 if baseline y is LDA, 2 if GGA)
            feature_list (FeatureList): FeatureList object
                (ciderpress.models.transform_data.FeatureList) containing
                the features to pass to the model.
            desc_order (list): indexes the descriptors before passing them
                to the features_list
            const: constant term to add to the spline
            desc_version ('c'): Version of the descriptors used ('a' or 'c')
            a0, fac_mul, amin: parameters passed to CIDER feature generator.
                (see ciderpress.dft.get_gaussian_grid_a/c)
        """
        self.scale = scale
        self.nterms = len(self.scale)
        self.ind_sets = ind_sets
        self.spline_grids = spline_grids
        self.coeff_sets = coeff_sets
        self.xed_y_converter = xed_y_converter
        self.feature_list = feature_list
        self.nfeat = feature_list.nfeat
        self.desc_order = desc_order
        self.const = const
        self.desc_version = desc_version
        self.a0 = a0
        self.fac_mul = fac_mul
        self.amin = amin
        self.args = args
        if self.desc_version in ["b", "d"] and self.args is not None:
            self.vvmul = args.gg_vvmul

    def _xedy(self, y, x, code):
        if self.xed_y_converter[-1] == 1:
            return self.xed_y_converter[code](y, x[:, 0])
        elif self.xed_y_converter[-1] == 2:
            return self.xed_y_converter[code](y, x[:, 0], x[:, 1])
        else:
            return self.xed_y_converter[code](y)

    def xed_to_y(self, y, x):
        return self._xedy(y, x, 0)

    def y_to_xed(self, y, x):
        return self._xedy(y, x, 1)

    def get_descriptors(self, x):
        return self.feature_list(x[:, self.desc_order])

    def predict_from_desc(self, X, vec_eval=False, max_order=3, min_order=0):
        res = np.zeros(X.shape[0]) + self.const
        if vec_eval:
            dres = np.zeros(X.shape)
        for t in range(self.nterms):
            if (not isinstance(self.ind_sets[t], int)) and len(
                self.ind_sets[t]
            ) < min_order:
                continue
            if isinstance(self.ind_sets[t], int) or len(self.ind_sets[t]) <= max_order:
                ind_set = self.ind_sets[t]
                if vec_eval:
                    y, dy = get_vec_eval(
                        self.spline_grids[t],
                        self.coeff_sets[t],
                        X[:, ind_set],
                        len(ind_set),
                    )
                    res += y * self.scale[t]
                    dres[:, ind_set] += dy * self.scale[t]
                else:
                    res += (
                        get_cubic(
                            self.spline_grids[t],
                            self.coeff_sets[t],
                            X[:, ind_set],
                            len(ind_set),
                        )
                        * self.scale[t]
                    )
        if vec_eval:
            return res, dres
        else:
            return res

    def predict(self, X, vec_eval=False):
        desc = self.get_descriptors(X)
        F = self.predict_from_desc(desc, vec_eval=vec_eval)
        if vec_eval:
            F = F[0]
        return self.y_to_xed(F, X)

    def to_dict(self):
        return {
            "scale": np.array(self.scale),
            "ind_sets": self.ind_sets,
            "spline_grids": self.spline_grids,
            "coeff_sets": self.coeff_sets,
            "xed_y_converter": self.xed_y_converter,
            "feature_list": self.feature_list,
            "desc_order": self.desc_order,
            "const": self.const,
            "desc_version": self.desc_version,
            "a0": self.a0,
            "fac_mul": self.fac_mul,
            "amin": self.amin,
            "args": self.args,
        }

    @classmethod
    def from_dict(cls, state_dict):
        return cls(
            state_dict["scale"],
            state_dict["ind_sets"],
            state_dict["spline_grids"],
            state_dict["coeff_sets"],
            state_dict["xed_y_converter"],
            state_dict["feature_list"],
            state_dict["desc_order"],
            const=state_dict["const"],
            desc_version=state_dict["desc_version"],
            a0=state_dict["a0"],
            fac_mul=state_dict["fac_mul"],
            amin=state_dict["amin"],
            args=state_dict["args"],
        )

    def dump(self, fname):
        """
        Save the Evaluator to a file name fname as yaml format.
        """
        state_dict = self.to_dict()
        with open(fname, "w") as f:
            yaml.dump(state_dict, f)

    @classmethod
    def load(cls, fname):
        """
        Load an instance of this class from yaml
        """
        with open(fname, "r") as f:
            state_dict = yaml.load(f, Loader=yaml.CLoader)
        return cls.from_dict(state_dict)


class MLFunctional(ABC):
    """
    Abstract base class for Machine-Learned functionals
    to be evaluated using ciderpress.dft.NLNumInt
    """

    def get_F(self, X):
        """
        Get exchange enhancement factor.
        """
        return self.get_F_and_derivative(X)[0]

    def get_derivative(self, X):
        """
        Get derivative of exchange enhancement factor
        with respect to descriptors.
        """
        return self.get_F_and_derivative(X)[1]

    @abstractmethod
    def get_F_and_derivative(self, X):
        """
        Get exchange enhancement factor and its derivative
        with respect to descriptors.
        """


#######################################################
# Semi-local functionals and simple helper functions. #
#######################################################


def identity(x):
    return x.copy()


def square(x):
    return x**2


def single(x):
    return np.ones(x.shape)


def extract_kernel_components(kernel):
    """
    For an sklearn Kernel object composed of
    three kernels of the form:
    kernel = (const * comp) + (noise),
    extract the three components
    """
    return kernel.k1.k1, kernel.k1.k2, kernel.k2


class Descriptor:
    """
    Old version of the Descriptor class (replaced with
    the models.transform_data module). Used in Semi-local
    functional examples below.
    """

    def __init__(self, code, transform=identity, transform_deriv=single, mul=1.0):
        self._transform = transform
        self._transform_deriv = transform_deriv
        self.code = code
        self.mul = mul

    def transform_descriptor(self, desc, deriv=0):
        if deriv == 0:
            return self._transform(desc[self.code])
        else:
            return self._transform(desc[self.code]), self._transform_deriv(
                desc[self.code]
            )


kappa = 0.804
mu = 0.2195149727645171


class PBEFunctional(MLFunctional):
    """
    PBE Exchange functional using the MLFunctional class
    """

    def __init__(self):
        self.desc_list = [Descriptor(1)]
        self.y_to_f_mul = None

    def get_F(self, X):
        p = X.flatten() ** 2
        return 1 + kappa - kappa / (1 + mu * p / kappa)

    def get_derivative(self, X):
        p = X.flatten() ** 2
        return (mu / (1 + mu * p / kappa) ** 2).reshape(-1, 1)


class SCANFunctional(MLFunctional):
    """
    SCAN Exchange functional using the MLFunctional class
    """

    def __init__(self):
        self.desc_list = [Descriptor(1), Descriptor(2)]
        self.y_to_f_mul = None

    def get_F(self, X):
        p = X[:, 0] ** 2
        s = X[:, 0]
        alpha = X[:, 1]
        muak = 10.0 / 81
        k1 = 0.065
        b2 = np.sqrt(5913 / 405000)
        b1 = (511 / 13500) / (2 * b2)
        b3 = 0.5
        b4 = muak**2 / k1 - 1606 / 18225 - b1**2
        h0 = 1.174
        a1 = 4.9479
        c1 = 0.667
        c2 = 0.8
        dx = 1.24
        tmp1 = muak * p
        tmp2 = 1 + b4 * p / muak * np.exp(-np.abs(b4) * p / muak)
        tmp3 = b1 * p + b2 * (1 - alpha) * np.exp(-b3 * (1 - alpha) ** 2)
        x = tmp1 * tmp2 + tmp3**2
        h1 = 1 + k1 - k1 / (1 + x / k1)
        gx = 1 - np.exp(-a1 / np.sqrt(s + 1e-9))
        -a1 / 4 * (s + 1e-9) ** (-2.5) * np.exp(-a1 / np.sqrt(s + 1e-9))
        fx = np.exp(-c1 * alpha / (1 - alpha)) * (alpha < 1) - dx * np.exp(
            c2 / (1 - alpha)
        ) * (alpha > 1)
        fx[np.isnan(fx)] = 0
        assert not np.isnan(fx).any()
        Fscan = gx * (h1 + fx * (h0 - h1))
        return Fscan

    def get_derivative(self, X):
        p = X[:, 0] ** 2
        s = X[:, 0]
        alpha = X[:, 1]
        muak = 10.0 / 81
        k1 = 0.065
        b2 = np.sqrt(5913 / 405000)
        b1 = (511 / 13500) / (2 * b2)
        b3 = 0.5
        b4 = muak**2 / k1 - 1606 / 18225 - b1**2
        h0 = 1.174
        a1 = 4.9479
        c1 = 0.667
        c2 = 0.8
        dx = 1.24
        tmp1 = muak * p
        tmp2 = 1 + b4 * p / muak * np.exp(-np.abs(b4) * p / muak)
        tmp3 = b1 * p + b2 * (1 - alpha) * np.exp(-b3 * (1 - alpha) ** 2)
        x = tmp1 * tmp2 + tmp3**2
        h1 = 1 + k1 - k1 / (1 + x / k1)
        gx = 1 - np.exp(-a1 / np.sqrt(s + 1e-9))
        dgdp = -a1 / 4 * (s + 1e-9) ** (-2.5) * np.exp(-a1 / np.sqrt(s + 1e-9))
        fx = np.exp(-c1 * alpha / (1 - alpha)) * (alpha < 1) - dx * np.exp(
            c2 / (1 - alpha)
        ) * (alpha > 1)
        fx[np.isnan(fx)] = 0
        assert not np.isnan(fx).any()
        gx * (h1 + fx * (h0 - h1))
        dxdp = (
            muak * tmp2
            + tmp1
            * (
                b4 / muak * np.exp(-np.abs(b4) * p / muak)
                - b4 * np.abs(b4) * p / muak**2 * np.exp(-np.abs(b4) * p / muak)
            )
            + 2 * tmp3 * b1
        )
        dxda = (
            2
            * tmp3
            * (
                -b2 * np.exp(-b3 * (1 - alpha) ** 2)
                + 2 * b2 * b3 * (1 - alpha) ** 2 * np.exp(-b3 * (1 - alpha) ** 2)
            )
        )
        dhdx = 1 / (1 + x / k1) ** 2
        dhdp = dhdx * dxdp
        dhda = dhdx * dxda
        dfda = (-c1 * alpha / (1 - alpha) ** 2 - c1 / (1 - alpha)) * np.exp(
            -c1 * alpha / (1 - alpha)
        ) * (alpha < 1) - dx * c2 / (1 - alpha) ** 2 * np.exp(c2 / (1 - alpha)) * (
            alpha > 1
        )
        dfda[np.isnan(dfda)] = 0

        dFdp = dgdp * (h1 + fx * (h0 - h1)) + gx * (1 - fx) * dhdp
        dFda = gx * (dhda - fx * dhda + dfda * (h0 - h1))
        return np.array([dFdp, dFda]).T


#########################
# GP-based functionals. #
#########################


class GPFunctional(MLFunctional):
    # TODO: This setup currently assumes that the gp directly
    # predict F_X - 1. This will not always be the case.

    def __init__(self, gpr):
        # Assumes kernel_ is (const * rbf) + noise
        msg = """
        GPFunctional is provided as a reference only. Its functional
        derivatives are buggy, so please do not use it for practical
        calculations.

        gpr is a ciderpress.models.gp.DFTGPR object
        """
        import warnings

        warnings.warn(msg)
        from sklearn.gaussian_process.kernels import RBF

        cov = gpr.gp.kernel_.k1
        self.alpha_ = cov.k1.constant_value * gpr.gp.alpha_
        self.kernel = RBF(length_scale=cov.k2.length_scale)
        self.X_train_ = gpr.gp.X_train_[:, 1:]
        self.feature_list = gpr.feature_list
        self.nfeat = self.feature_list.nfeat
        self.desc_version = gpr.args.version[:1]
        self.a0 = gpr.args.gg_a0
        self.amin = gpr.args.gg_amin
        self.fac_mul = gpr.args.gg_facmul
        self.vvmul = gpr.args.gg_vvmul
        self.desc_order = gpr.args.desc_order
        self.fx_baseline = gpr.xed_y_converter[2]
        self.fxb_num = gpr.xed_y_converter[3]
        self.args = gpr.args

    def get_F_and_derivative(self, X):
        rho = X[0]
        mat = np.zeros((self.nfeat, X.shape[1]))
        self.feature_list.fill_vals_(mat, X)

        # X has shape n_test, n_desc
        # X_train_ has shape n_train, n_desc
        k = self.kernel(mat.T, self.X_train_)
        F = k.dot(self.alpha_)
        k * self.alpha_
        # shape n_test, n_desc
        kaxt = np.dot(k * self.alpha_, self.X_train_)
        np.dot(k, self.alpha_)
        dF = kaxt - mat.T * F.reshape(-1, 1)
        dF /= self.kernel.length_scale**2
        dFddesc = np.zeros(X.shape).T
        self.feature_list.fill_derivs_(dFddesc.T, dF.T, X)

        highcut = 1e-3
        lowcut = 1e-6
        rhocut = np.maximum(rho[rho < highcut], lowcut)
        xcut = np.log(rhocut / lowcut) / np.log(highcut / lowcut)
        F[rho < highcut] *= 0.5 * (1 - np.cos(np.pi * xcut))
        dFddesc[rho < highcut, :] *= 0.5 * (1 - np.cos(np.pi * xcut[:, np.newaxis]))
        dFddesc[rho < lowcut, :] = 0

        if self.fxb_num == 1:
            chfx = 1
        elif self.fxb_num == 2:
            chfx, dchfx = self.fx_baseline(X[1])
            dFddesc[:, 1] += dchfx
        else:
            raise ValueError("Unsupported basline fx order.")
        F += chfx

        F[rho < 1e-9] = 0
        dFddesc[rho < 1e-9, :] = 0

        return F, dFddesc

    def get_F(self, X):
        return self.get_F_and_derivative(X)[0]

    def get_derivative(self, X):
        return self.get_F_and_derivative(X)[1]


class NormGPFunctional(MLFunctional, Evaluator):
    """
    MLFunctional to evaluate spline-interpolated GP
    functionals using the utilities in the Evaluator class.
    """

    def __init__(self, *args, **kwargs):
        """
        Same arguments as Evaluator.
        """
        Evaluator.__init__(self, *args, **kwargs)
        self.fxb_num = self.xed_y_converter[3]
        self.fx_baseline = self.xed_y_converter[2]

    def get_F_and_derivative(self, X):
        rho = X[0]
        mat = np.zeros((self.nfeat, X.shape[1]))
        self.feature_list.fill_vals_(mat, X)
        F, dF = self.predict_from_desc(mat.T, vec_eval=True)
        dFddesc = np.zeros(X.shape).T
        self.feature_list.fill_derivs_(dFddesc.T, dF.T, X)

        highcut = 1e-8
        lowcut = 1e-9
        rhocut = np.maximum(rho[rho < highcut], lowcut)
        xcut = np.log(rhocut / lowcut) / np.log(highcut / lowcut)
        F[rho < highcut] *= 0.5 * (1 - np.cos(np.pi * xcut))
        dFddesc[rho < highcut, :] *= 0.5 * (1 - np.cos(np.pi * xcut[:, np.newaxis]))
        dFddesc[rho < lowcut, :] = 0

        if self.fxb_num == 1:
            chfx = 1
        elif self.fxb_num == 2:
            chfx, dchfx = self.fx_baseline(X[1])
            dFddesc[:, 1] += dchfx
        else:
            raise ValueError("Unsupported baseline fx order.")
        F += chfx

        if rho is not None:
            F[rho < 1e-9] = 0
            dFddesc[rho < 1e-9, :] = 0

        return F, dFddesc



#!/usr/bin/env python
# CiderPress: Machine-learning based density functional theory calculations
# Copyright (C) 2024 The President and Fellows of Harvard College
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# Author: Kyle Bystrom <kylebystrom@gmail.com>
#

import ctypes
import time

import numpy as np
from pyscf import lib
from pyscf.dft.numint import _dot_ao_ao, _dot_ao_dm, _scale_ao, eval_ao
from pyscf.gto.eval_gto import BLKSIZE, _get_intor_and_comp, make_screen_index
from pyscf.gto.mole import ANG_OF, ATOM_OF, PTR_EXP

from ciderpress.dft.plans import SADMPlan, SDMXFullPlan, SDMXIntPlan, SDMXPlan
from ciderpress.dft.settings import SADMSettings, SDMXFullSettings, SDMXSettings
from ciderpress.dft.sph_harm_coeff import get_deriv_ylm_coeff
from ciderpress.lib import load_library as load_cider_library

libcider = load_cider_library("libmcider")


def _get_ylm_atom_loc(mol):
    ylm_atom_loc = np.zeros(mol.natm + 1, dtype=np.int32)
    for ia in range(mol.natm):
        lmax = np.max(mol._bas[mol._bas[:, ATOM_OF] == ia, ANG_OF]) + 1
        ylm_atom_loc[ia + 1] = ylm_atom_loc[ia] + lmax * lmax
    return ylm_atom_loc


def eval_conv_shells(
    feval,
    plan,
    mol,
    coords,
    shls_slice=None,
    non0tab=None,
    ao_loc=None,
    cutoff=None,
    out=None,
):
    if isinstance(plan, tuple):
        alphas, alpha_norms, itype = plan
    else:
        alphas = plan.alphas
        alpha_norms = plan.alpha_norms
        itype = plan.settings.integral_type
    if non0tab is not None:
        if (non0tab == 0).any():
            # TODO implement some sort of screening later
            raise NotImplementedError
    eval_name, comp = _get_intor_and_comp(mol, feval)
    if comp not in [1, 4]:
        raise NotImplementedError
    comp = len(alphas)
    if eval_name == "GTOval_sph_deriv0":
        ncpa = 1
        if itype == "gauss_diff":
            contract_fn = getattr(libcider, "SDMXcontract_smooth0")
        elif itype == "gauss_r2":
            contract_fn = getattr(libcider, "SDMXcontract_rsq0")
        else:
            raise ValueError
        eval_fn = getattr(libcider, "SDMXrad_eval_grid")
    elif eval_name == "GTOval_sph_deriv1":
        ncpa = 2
        if itype == "gauss_diff":
            contract_fn = getattr(libcider, "SDMXcontract_smooth1")
        elif itype == "gauss_r2":
            contract_fn = getattr(libcider, "SDMXcontract_rsq1")
        else:
            raise ValueError
        eval_fn = getattr(libcider, "SDMXrad_eval_grid_deriv1")
    else:
        raise NotImplementedError
    comp *= ncpa
    drv = getattr(libcider, "SDMXeval_rad_loop")
    atm = np.asarray(mol._atm, dtype=np.int32, order="C")
    bas = np.asarray(mol._bas, dtype=np.int32, order="C")
    env = np.asarray(mol._env, dtype=np.double, order="C")
    coords = np.asarray(coords, dtype=np.double, order="F")
    natm = atm.shape[0]
    nbas = bas.shape[0]
    ngrids = coords.shape[0]

    if shls_slice is None:
        shls_slice = (0, nbas)
    if "spinor" in eval_name:
        raise NotImplementedError
        # ao = np.ndarray((2, comp, nao, ngrids), dtype=np.complex128,
        #                 buffer=out).transpose(0, 1, 3, 2)
    else:
        ao = np.ndarray((comp, nbas, ngrids), buffer=out).transpose(0, 2, 1)

    if non0tab is None:
        if cutoff is None:
            non0tab = np.ones(((ngrids + BLKSIZE - 1) // BLKSIZE, nbas), dtype=np.uint8)
        else:
            non0tab = make_screen_index(mol, coords, shls_slice, cutoff)
    time.monotonic()
    drv(
        eval_fn,
        contract_fn,
        ctypes.c_double(1),
        ctypes.c_int(ngrids),
        (ctypes.c_int * 2)(1, ncpa),
        (ctypes.c_int * 2)(*shls_slice),
        ao.ctypes.data_as(ctypes.c_void_p),
        coords.ctypes.data_as(ctypes.c_void_p),
        non0tab.ctypes.data_as(ctypes.c_void_p),
        atm.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(natm),
        bas.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(nbas),
        env.ctypes.data_as(ctypes.c_void_p),
        alphas.ctypes.data_as(ctypes.c_void_p),
        alpha_norms.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(alphas.size),
    )
    time.monotonic()
    # print("Get shells", t1 - t0)
    if comp == 1:
        if "spinor" in eval_name:
            ao = ao[:, 0]
        else:
            ao = ao[0]
    return ao


def eval_conv_sh(
    plan,
    mol,
    coords,
    deriv=0,
    shls_slice=None,
    non0tab=None,
    cutoff=None,
    out=None,
    verbose=None,
):
    assert deriv in [0, 1]
    if mol.cart:
        feval = "GTOval_cart_deriv%d" % deriv
    else:
        feval = "GTOval_sph_deriv%d" % deriv
    return eval_conv_shells(
        feval, plan, mol, coords, shls_slice, non0tab, cutoff=cutoff, out=out
    )


class PySCFSDMXInitializer:
    def __init__(self, settings, **kwargs):
        self.settings = settings
        self.kwargs = kwargs

    def initialize_sdmx_generator(self, mol, nspin):
        return EXXSphGenerator.from_settings_and_mol(
            self.settings, nspin, mol, **self.kwargs
        )


class EXXSphGenerator:
    fast = True

    def __init__(self, plan):
        if plan.fit_metric != "ovlp":
            raise NotImplementedError
        self.plan = plan
        self._ao_buf = None
        self._cao_buf = None
        self._ylm_buf = None
        self._shl_c0 = None
        self._cached_ao_data = None

    @property
    def has_l1(self):
        return self.plan.settings.n1terms > 0

    def reset_buffers(self):
        self._ao_buf = None
        self._cao_buf = None
        self._ylm_buf = None
        self._shl_c0 = None
        self._cached_ao_data = None

    @classmethod
    def from_settings_and_mol(
        cls, settings, nspin, mol, alpha0=None, lambd=None, nalpha=None, lowmem=False
    ):
        if alpha0 is None:
            min_exp = np.min(mol._env[mol._bas[:, PTR_EXP]])
            max_width = 1.0 / np.sqrt(2 * min_exp)
            coords = mol.atom_coords(unit="Bohr")
            dist = np.linalg.norm(coords[:, None, :] - coords[None, :, :], axis=2)
            max_dist = np.max(dist)
            max_wpd = 4 * max_dist + 4 * max_width
            # max_wpd = 2 * max_dist + 2 * max_width
            # max_wpd *= 2
            alpha0 = 1.0 / (2 * max_wpd**2)
        if lambd is None:
            lambd = 1.8
        if nalpha is None:
            max_exp = np.max(mol._env[mol._bas[:, PTR_EXP]])
            # max_exp = min(max_exp, 300)
            nalpha = 1 + int(1 + np.log(max_exp / alpha0) / np.log(lambd))
        if isinstance(settings, SDMXFullSettings):
            plan = SDMXFullPlan(settings, nspin, alpha0, lambd, nalpha)
            # plan = SDMXIntPlan(settings, nspin, alpha0, lambd, nalpha)
        elif isinstance(settings, SDMXSettings):
            plan = SDMXPlan(settings, nspin, alpha0, lambd, nalpha)
        elif isinstance(settings, SADMSettings):
            plan = SADMPlan(settings, nspin, alpha0, lambd, nalpha, fit_metric="ovlp")
        else:
            raise ValueError("Invalid type {}".format(type(settings)))
        return cls(plan)

    def __call__(self, *args, **kwargs):
        """
        Evaluate the features. See get_features.
        """
        return self.get_features(*args, **kwargs)

    def _get_fit_mats(self):
        if self.plan.has_coul_list:
            fit_mats = self.plan.fit_matrices
        else:
            fit_mats = [self.plan.fit_matrix]
        return fit_mats

    @property
    def deriv(self):
        return 1 if self.has_l1 else 0

    def get_extra_ao(self, mol):
        ymem = _get_ylm_atom_loc(mol)[-1]
        cmem = (1 + 6 * self.deriv) * mol.nbas
        bmem = (1 + self.deriv) * mol.nbas * self.plan.nalpha
        return ymem + cmem + bmem

    def get_ao(self, mol, coords, non0tab=None, cutoff=None, save_buf=True):
        # TODO use buffers
        return eval_ao(mol, coords, deriv=0, non0tab=non0tab, cutoff=cutoff, out=None)

    def get_cao(self, mol, coords, non0tab=None, cutoff=None, save_buf=True):
        caobuf_size = mol.nbas * coords.shape[0] * self.plan.nalpha
        if self.deriv > 0:
            caobuf_size *= 2
        if self._cao_buf is not None and self._cao_buf.size >= caobuf_size:
            buf = self._cao_buf
        else:
            buf = np.empty(caobuf_size)
        res = eval_conv_sh(
            self.plan,
            mol,
            coords,
            deriv=self.deriv,
            non0tab=non0tab,
            cutoff=cutoff,
            out=buf,
        )
        if save_buf:
            self._cao_buf = buf
        return res

    def _contract_ao_to_bas_single_(
        self,
        mol,
        b0,
        ylm,
        c0,
        shls_slice,
        ao_loc,
        ylm_atom_loc,
        coords=None,
        atom_coords=None,
        bwd=False,
    ):
        args = [
            ctypes.c_int(b0.shape[-1]),
            b0.ctypes.data_as(ctypes.c_void_p),
            ylm.ctypes.data_as(ctypes.c_void_p),
            c0.ctypes.data_as(ctypes.c_void_p),
            (ctypes.c_int * 2)(*shls_slice),
            ao_loc.ctypes.data_as(ctypes.c_void_p),
            ylm_atom_loc.ctypes.data_as(ctypes.c_void_p),
            mol._atm.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(mol.natm),
            mol._bas.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(mol.nbas),
            mol._env.ctypes.data_as(ctypes.c_void_p),
        ]
        if coords is not None:
            assert coords.ndim == 1 and coords.flags.c_contiguous
            assert atom_coords.ndim == 1 and atom_coords.flags.c_contiguous
            args.append(coords.ctypes.data_as(ctypes.c_void_p))
            args.append(atom_coords.ctypes.data_as(ctypes.c_void_p))
            if bwd:
                libcider.SDMXcontract_ao_to_bas_grid_bwd(*args)
            else:
                libcider.SDMXcontract_ao_to_bas_grid(*args)
        else:
            if bwd:
                libcider.SDMXcontract_ao_to_bas_bwd(*args)
            else:
                libcider.SDMXcontract_ao_to_bas(*args)

    def _get_ylm(self, mol, coords, ylm_atom_loc=None, savebuf=True):
        ngrids = coords.shape[0]
        ncpa = 1 + 3 * self.deriv
        gaunt_lmax = np.max(mol._bas[:, ANG_OF])
        if ylm_atom_loc is None:
            ylm_atom_loc = _get_ylm_atom_loc(mol)
        ybufsize = ncpa * ylm_atom_loc[-1] * ngrids
        if self._ylm_buf is not None and self._ylm_buf.size >= ybufsize:
            buf = self._ylm_buf
        else:
            buf = np.empty(ybufsize)
        ylm = np.ndarray((ncpa, ylm_atom_loc[-1], ngrids), buffer=buf, order="C")
        atom_coords = np.ascontiguousarray(mol.atom_coords(unit="Bohr"))
        libcider.SDMXylm_loop(
            ctypes.c_int(coords.shape[0]),
            ylm.ctypes.data_as(ctypes.c_void_p),
            coords.ctypes.data_as(ctypes.c_void_p),
            ylm_atom_loc.ctypes.data_as(ctypes.c_void_p),
            atom_coords.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(mol.natm),
        )
        if ncpa > 1:
            gaunt_coeff = get_deriv_ylm_coeff(gaunt_lmax)
            libcider.SDMXylm_grad(
                ctypes.c_int(coords.shape[0]),
                ylm.ctypes.data_as(ctypes.c_void_p),
                gaunt_coeff.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(gaunt_coeff.shape[1]),
                ylm_atom_loc.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(mol.natm),
            )
        libcider.SDMXylm_yzx2xyz(
            ctypes.c_int(coords.shape[0]),
            ctypes.c_int(ncpa),
            ylm.ctypes.data_as(ctypes.c_void_p),
            ylm_atom_loc.ctypes.data_as(ctypes.c_void_p),
            ctypes.c_int(mol.natm),
        )
        if savebuf:
            self._ylm_buf = buf
        else:
            self._ylm_buf = None
        return ylm

    def _contract_ao_to_bas_helper(
        self, mol, b0, c0, shls_slice, ao_loc, coords, ylm=None, bwd=False
    ):
        coords = np.asfortranarray(coords)
        if ylm is None:
            ylm = self._get_ylm(mol, coords)
        ylm_atom_loc = _get_ylm_atom_loc(mol)
        if self.deriv == 1:
            atom_coords = np.asfortranarray(mol.atom_coords(unit="Bohr"))
            args = [
                ctypes.c_int(b0.shape[-1]),
                b0.ctypes.data_as(ctypes.c_void_p),
                ylm.ctypes.data_as(ctypes.c_void_p),
                c0.ctypes.data_as(ctypes.c_void_p),
                (ctypes.c_int * 2)(*shls_slice),
                ao_loc.ctypes.data_as(ctypes.c_void_p),
                ylm_atom_loc.ctypes.data_as(ctypes.c_void_p),
                mol._atm.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(mol.natm),
                mol._bas.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(mol.nbas),
                mol._env.ctypes.data_as(ctypes.c_void_p),
                coords.ctypes.data_as(ctypes.c_void_p),
                atom_coords.ctypes.data_as(ctypes.c_void_p),
            ]
            if bwd:
                fn = libcider.SDMXcontract_ao_to_bas_l1_bwd
            else:
                fn = libcider.SDMXcontract_ao_to_bas_l1
            fn(*args)
        else:
            self._contract_ao_to_bas_single_(
                mol,
                b0[0],
                ylm[0],
                c0,
                shls_slice,
                ao_loc,
                ylm_atom_loc,
                coords=None,
                bwd=bwd,
            )
        """
        if self.deriv > 0:
            atom_coords = np.asfortranarray(mol.atom_coords(unit='Bohr'))
            self._contract_ao_to_bas_single_(
                mol, b0[1], ylm[1], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=None, bwd=bwd
            )
            self._contract_ao_to_bas_single_(
                mol, b0[2], ylm[2], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=None, bwd=bwd
            )
            self._contract_ao_to_bas_single_(
                mol, b0[3], ylm[3], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=None, bwd=bwd
            )
            self._contract_ao_to_bas_single_(
                mol, b0[4], ylm[0], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=coords[:, 0],
                atom_coords=atom_coords[:, 0], bwd=bwd
            )
            self._contract_ao_to_bas_single_(
                mol, b0[5], ylm[0], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=coords[:, 1],
                atom_coords=atom_coords[:, 1], bwd=bwd
            )
            self._contract_ao_to_bas_single_(
                mol, b0[6], ylm[0], c0, shls_slice, ao_loc,
                ylm_atom_loc, coords=coords[:, 2],
                atom_coords=atom_coords[:, 2], bwd=bwd
            )
        """

    def _contract_ao_to_bas(self, mol, c0, shls_slice, ao_loc, coords, ylm=None):
        ngrids = coords.shape[0]
        b0 = np.empty((1 + 6 * self.deriv, mol.nbas, ngrids))
        self._contract_ao_to_bas_helper(
            mol,
            b0,
            c0,
            shls_slice,
            ao_loc,
            coords,
            ylm=ylm,
            bwd=False,
        )
        return b0

    def _contract_ao_to_bas_bwd(self, mol, b0, shls_slice, ao_loc, coords, ylm=None):
        ngrids = coords.shape[0]
        c0 = np.zeros((mol.nao_nr(), ngrids))
        self._contract_ao_to_bas_helper(
            mol,
            b0,
            c0,
            shls_slice,
            ao_loc,
            coords,
            ylm=ylm,
            bwd=True,
        )
        return c0.T

    def _eval_crho_potential(
        self, mol, coords, cao, tmp2, shls_slice, ao_loc, ylm=None
    ):
        ngrids = coords.shape[0]
        nalpha = self.plan.nalpha
        b0 = np.empty((1 + 6 * self.deriv, mol.nbas, ngrids)).transpose(0, 2, 1)
        if self.deriv == 0:
            b0[0] = _scale_ao(cao[:nalpha], tmp2[0])
        else:
            libcider.contract_shl_to_alpha_l1_bwd(
                ctypes.c_int(coords.shape[0]),
                ctypes.c_int(nalpha),
                ctypes.c_int(cao.shape[-1]),
                tmp2.ctypes.data_as(ctypes.c_void_p),
                b0.ctypes.data_as(ctypes.c_void_p),
                cao.ctypes.data_as(ctypes.c_void_p),
            )
        # if self.deriv > 0:
        #     b0[1] = _scale_ao(cao[:nalpha], tmp2[1])
        #     b0[2] = _scale_ao(cao[:nalpha], tmp2[2])
        #     b0[3] = _scale_ao(cao[:nalpha], tmp2[3])
        #     b0[4] = _scale_ao(cao[nalpha:2*nalpha], tmp2[1])
        #     b0[5] = _scale_ao(cao[nalpha:2*nalpha], tmp2[2])
        #     b0[6] = _scale_ao(cao[nalpha:2*nalpha], tmp2[3])
        return self._contract_ao_to_bas_bwd(
            mol, b0.transpose(0, 2, 1), shls_slice, ao_loc, coords, ylm=ylm
        )

    def get_features(
        self,
        dms,
        mol,
        coords,
        non0tab=None,
        cutoff=None,
        save_buf=False,
        ao=None,
        cao=None,
    ):
        coords = np.asfortranarray(coords)
        ndim = dms.ndim
        if ndim == 2:
            dms = dms[None, :, :]
            n_out = 1
        elif ndim == 3:
            n_out = dms.shape[0]
        else:
            raise ValueError
        has_l1 = self.has_l1
        ngrids = coords.shape[0]
        nalpha = self.plan.nalpha
        ncpa = 4 if has_l1 else 1
        n0 = self.plan.num_l0_feat
        nfeat = self.plan.settings.nfeat
        # TODO might want to account for these when computing
        # memory footprint, though they are linear scaling in
        # size and therefore less of an issue.
        tmp = np.empty((ncpa, nalpha, ngrids))
        if isinstance(self.plan, SDMXIntPlan):
            l0tmp2 = np.empty((n_out, nalpha, ngrids))
            l1tmp2 = np.empty((n_out, 3, nalpha, ngrids))
        else:
            l0tmp2 = np.empty((n_out, n0, nalpha, ngrids))
            l1tmp2 = np.empty((n_out, nfeat - n0, 3, nalpha, ngrids))
        out = np.empty((n_out, nfeat, ngrids))
        time.monotonic()
        if ao is None:
            ao = self.get_ao(
                mol, coords, non0tab=non0tab, cutoff=cutoff, save_buf=save_buf
            )
        if cao is None:
            cao = self.get_cao(
                mol, coords, non0tab=non0tab, cutoff=cutoff, save_buf=save_buf
            )
        time.monotonic()
        shls_slice = (0, mol.nbas)
        ao_loc = mol.ao_loc_nr()
        ylm = self._get_ylm(mol, coords, savebuf=True)
        for idm, dm in enumerate(dms):
            c0 = _dot_ao_dm(mol, ao, dm, None, shls_slice, ao_loc)
            time.monotonic()
            b0 = self._contract_ao_to_bas(mol, c0, shls_slice, ao_loc, coords, ylm=ylm)
            time.monotonic()
            if has_l1:
                assert tmp.flags.c_contiguous
                assert b0.flags.c_contiguous
                libcider.contract_shl_to_alpha_l1(
                    ctypes.c_int(coords.shape[0]),
                    ctypes.c_int(nalpha),
                    ctypes.c_int(cao.shape[-1]),
                    tmp.ctypes.data_as(ctypes.c_void_p),
                    b0.ctypes.data_as(ctypes.c_void_p),
                    cao.ctypes.data_as(ctypes.c_void_p),
                )
            else:
                for ialpha in range(nalpha):
                    tmp[0, ialpha] = lib.einsum(
                        "bg,bg->g", b0[0], cao[0 * nalpha + ialpha].T
                    )
            time.monotonic()
            self.plan.get_features(
                tmp, out=out[idm], l0tmp=l0tmp2[idm], l1tmp=l1tmp2[idm]
            )
            time.monotonic()
        # print('times', t1 - t0, t2 - t1, t3 - t2, torb2 - torb)
        # print(out.mean(axis=-1))
        # exit()
        if ndim == 2:
            out = out[0]
        self._cached_ao_data = (mol, ao, cao, l0tmp2, l1tmp2, coords, ylm)
        return out

    def get_vxc_(self, vxc_mat, vxc_grid):
        if self._cached_ao_data is None:
            raise RuntimeError("Must call get_features first")
        ndim = vxc_mat.ndim
        # nalpha = self.plan.nalpha
        # ngrids = vxc_grid.shape[-1]
        if ndim == 2:
            vxc_mat = vxc_mat[None, :, :]
            vxc_grid = vxc_grid[None, :]
            n_out = 1
        elif ndim == 3:
            n_out = vxc_mat.shape[0]
        else:
            raise ValueError
        # has_l1 = self.has_l1
        mol, ao, cao, alpha_terms, l1_alpha_terms, coords, ylm = self._cached_ao_data
        # ncpa = 1 + 3 * self.deriv
        # n0 = self.plan.num_l0_feat
        shls_slice = (0, mol.nbas)
        ao_loc = mol.ao_loc_nr()
        for idm in range(n_out):
            tmp2 = self.plan.get_vxc(
                vxc_grid[idm], alpha_terms[idm], l1_alpha_terms[idm]
            )
            tmp3 = self._eval_crho_potential(
                mol, coords, cao, tmp2, shls_slice, ao_loc, ylm=ylm
            )
            vxc_mat[idm] += _dot_ao_ao(mol, ao, tmp3, None, shls_slice, ao_loc, hermi=0)
        return vxc_mat
